#include "client.hpp"
#include "../buffer_operator/streamed_adder.hpp"
#include "../buffer_operator/line_consumer.hpp"
#include "../util/buffer_streamer.hpp"
#include "../util/parser.hpp"
#include "../logger/null.hpp"
#include "hostaddress.hpp"

namespace fz::http
{

client::client(thread_pool &pool, event_loop &loop, logger_interface &logger, tls_system_trust_store &trust_store, options opts)
	: fz::event_handler(loop)
	, logger::modularized(logger, "HTTP Client")
	, adder()
	, message_consumer(*this, 4096)
	, pool_(pool)
	, trust_store_(trust_store)
	, logger_(*this)
	, opts_(std::move(opts))
	, channel_(*this, 4*128*1024, 5, false)
{
}

client::~client()
{
	remove_handler();
}

void fz::http::client::enqueue_request(request &&request, response::handler &&handler)
{
	if (!fz::equal_insensitive_ascii(request.uri.scheme_, "https") && !fz::equal_insensitive_ascii(request.uri.scheme_, "http") && !request.uri.scheme_.empty()) {
		handler({400, {}, fz::sprintf("Unknown scheme [%s]", request.uri.scheme_), {}});
		return;
	}

	{
		fz::scoped_lock lock(mutex_);

		requests_.push_back(std::move(request));
		response_handlers_.push_back(std::move(handler));
	}

	process_queue();
}

int client::add_to_buffer()
{
	fz::scoped_lock lock(mutex_);

	if (requests_.empty())
		return EAGAIN;

	if (current_response_)
		// TODO: pipelining not supported yet.
		return EAGAIN;

	auto &req = current_request_ = std::move(requests_.front());

	if (req.uri.path_.empty())
		req.uri.path_ = "/";

	// Some servers do not expect Host to contain the "standard" port.
	if ((fz::equal_insensitive_ascii(req.uri.scheme_, "https") && req.uri.port_ != 443) || (fz::equal_insensitive_ascii(req.uri.scheme_, "http") && req.uri.port_ != 80))
		req.headers["Host"] = join_host_and_port(req.uri.host_, req.uri.port_);
	else
		req.headers["Host"] = req.uri.host_;

	req.headers["User-Agent"] = opts_.user_agent();
	req.headers["Connection"] = "close";

	if (req.body.size() > 0)
		req.headers["Content-length"] = std::to_string(req.body.size());

	for (const auto &d: opts_.default_request_headers()) {
		if (req.headers.find(d.first) == req.headers.end())
			req.headers[d.first] = d.second;
	}

	auto buffer = adder::get_buffer();
	util::buffer_streamer out(*buffer);

	out << req.method << " " << req.uri.get_request() << " HTTP/1.1\r\n";

	for (const auto &h: req.headers) {
		out << h.first << ": " << h.second << "\r\n";
	}

	out << "\r\n";
	out << req.body;

	logger_.log_u(logmsg::debug_debug, L"add_to_buffer()/request: %s***END", buffer->to_view());

	requests_.pop_front();

	message_consumer::reset();

	current_response_.emplace();
	current_response_handler_ = std::move(response_handlers_.front());
	response_handlers_.pop_front();

	return 0;
}

int client::process_message_start_line(std::string_view line)
{
	if (line.size() < 13 || !fz::equal_insensitive_ascii(line.substr(0, 7), std::string_view("HTTP/1."))) {
		logger_.log_u(logmsg::error, "Invalid status line");
		return EINVAL;
	}

	if (line[9]  < '1' || line[9]  > '5' ||
		line[10] < '0' || line[10] > '9' ||
		line[11] < '0' || line[11] > '9')
	{
		logger_.log_u(logmsg::error, "Invalid response code");
		return EINVAL;
	}

	current_response_->code = unsigned((line[9] - '0') * 100 + (line[10] - '0') * 10 + line[11] - '0');

	if (current_response_->code == 305) {
		logger_.log_u(logmsg::error, "Unsupported redirect");
		return EINVAL;
	}

	current_response_->reason = line.substr(13);

	return 0;
}

int client::process_message_header(std::string_view key, std::string_view value)
{
	// FIXME: Cookies must be handled differently, for now just ignore them.
	// See rfc7230§3.2.2
	if (fz::equal_insensitive_ascii(key, "Set-Cookie"))
		return 0;

	// We don't need to memorize this
	if (fz::equal_insensitive_ascii(key, "Transfer-Encoding"))
		return 0;

	auto &header = current_response_->headers[std::string(key)];
	if (header.empty())
		header = value;
	else
		header.append(", ").append(value);

	return 0;
}

int client::process_body_chunk(buffer_string_view chunk)
{
	auto space_left = opts_.max_body_size() - current_response_->body.size();

	if (chunk.size() > space_left) {
		logger_.log_u(logmsg::debug_warning, L"Message body size exceeds the maximum allowed size");

		chunk.remove_suffix(chunk.size() - space_left);
		current_response_->body.append(chunk.begin(), chunk.end());
		return process_end_of_message();
	}

	current_response_->body.append(chunk.begin(), chunk.end());

	return 0;
}

int client::process_end_of_message_headers()
{
	current_response_headers_have_been_parsed_ = true;

	if (opts_.follow_redirects() && current_response_->code >= 300 && current_response_->code < 305) {
		if (auto it = current_response_->headers.find("Location"); it != current_response_->headers.end()) {
			if (!opts_.redirects_limits() || redirection_num_ < opts_.redirects_limits()) {
				redirection_num_ += 1;

				auto new_uri = fz::uri(it->second);
				new_uri.resolve(current_request_.uri);
				current_request_.uri = std::move(new_uri);
				enqueue_request(std::move(current_request_), std::move(current_response_handler_));
				close(-int(current_response_->code));
				return EAGAIN;
			}
			else {
				logger_.log_u(logmsg::debug_warning, L"Redirection limit %d reached. Not following the redirection.", opts_.redirects_limits());
				return EINVAL;
			}
		}
		else {
			logger_.log_u(logmsg::debug_warning, L"Response code is %d, but new Location was not provided.", current_response_->code);
		}
	}

	redirection_num_ = 0;

	return 0;
}

int client::process_end_of_message()
{
	close(0);
	return EAGAIN;
}

void client::process_queue()
{
	fz::scoped_lock lock(mutex_);

	if (!socket_ && requests_.size() > 0) {
		auto &request = requests_.front();

		must_secure_socket_ = fz::equal_insensitive_ascii(request.uri.scheme_, "https");

		if (!request.uri.port_)
			request.uri.port_ =  must_secure_socket_ ? 443 : 80;

		logger_.log_u(logmsg::debug_debug, L"Connecting to %s", join_host_and_port(request.uri.host_, request.uri.port_));
		socket_ = std::make_unique<securable_socket>(event_loop_, static_cast<event_handler*>(this), std::make_unique<socket>(pool_, static_cast<event_handler *>(this)), logger::null);
		if (int error = socket_->connect(fz::to_native(fz::to_wstring_from_utf8(request.uri.host_)), request.uri.port_)) {
			event_handler::send_event<socket_event>(socket_->root(), socket_event_flag::connection, error);
			return;
		}
	}
}

void client::close(int err)
{
	if (err != 0) {
		current_response_.reset();

		if (err < 0)
			err = 0;
	}

	channel_.set_socket(socket_.get());
	channel_.shutdown(err);
}

void client::on_connection_error(int error)
{
	fz::scoped_lock lock(mutex_);

	auto request = std::move(requests_.front());
	auto handler = std::move(response_handlers_.front());
	response_handlers_.pop_front();
	requests_.pop_front();

	if (handler) {
		lock.unlock();

		handler({503, {}, fz::sprintf("Could not connect to host %. Reason: %s.", join_host_and_port(request.uri.host_, request.uri.port_), socket_error_description(error)), {}});

		lock.lock();
	}

	close(error);
}


void client::on_certificate_verification_event(tls_layer *tls, tls_session_info &info)
{
	bool is_trusted = opts_.cert_verifier() ? opts_.cert_verifier()(info) : info.system_trust();

	tls->set_verification_result(is_trusted);
	logger_.log_u(logmsg::debug_debug, L"Certificate is trusted: %s", is_trusted ? L"yes" : L"no");
}

void client::on_socket_event(socket_event_source *source, socket_event_flag type, int error)
{
	if (error) {
		logger_.log_u(logmsg::error, L"Error %d during connection", error);
		return on_connection_error(error);
	}

	if (type == socket_event_flag::connection) {
		if (must_secure_socket_) {
			if (!socket_->make_secure_client(tls_ver::v1_2, {}, &trust_store_, {}, {"http/1.1"})) {
				logger_.log_u(logmsg::error, L"socket_->make_secure_client() failed. Shutting down.");
				return on_connection_error(ESHUTDOWN);
			}

			must_secure_socket_ = false;
			return;
		}

		channel_.set_buffer_adder(this);
		channel_.set_buffer_consumer(this);
		channel_.set_socket(socket_.get());
		return;
	}

	logger_.log_u(logmsg::error,
				L"We got an unexpected socket_event. is_own_socket [%d], flag [%d], error [%d], from a layer [%d]",
				source->root() == socket_->root(), type, error, source != source->root());

	//assert(false && "We got an unexpected socket_event");
}

void client::on_channel_done_event(channel &, channel::error_type error)
{
	if (error) {
		auto msg = fz::sprintf("Channel closed with error from source %d. Reason: %s.", (int)error.source(), socket_error_description(error.error()));

		logger_.log_u(logmsg::error, L"%s", msg);

		if (current_response_ && !current_response_headers_have_been_parsed_)
			current_response_ = { 503, {}, msg, {} };
	}

	if (current_response_) {
		if (current_response_handler_)
			current_response_handler_(std::move(*current_response_));
	}

	current_response_.reset();
	current_response_headers_have_been_parsed_ = false;
	socket_.reset();
	process_queue();
}

void client::operator ()(const event_base &ev)
{
	fz::dispatch<
		socket_event,
		channel::done_event,
		certificate_verification_event
	>(ev, this,
		&client::on_socket_event,
		&client::on_channel_done_event,
		&client::on_certificate_verification_event
	);
}

}
