#ifndef FZ_HTTP_CLIENT_HPP
#define FZ_HTTP_CLIENT_HPP

#include <deque>
#include <libfilezilla/tls_system_trust_store.hpp>

#include "config.hpp"

#include "../logger/modularized.hpp"
#include "../tcp/client.hpp"
#include "../buffer_operator/adder.hpp"
#include "../buffer_operator/line_consumer.hpp"
#include "../util/options.hpp"
#include "../securable_socket.hpp"

#include "request.hpp"
#include "response.hpp"
#include "message_consumer.hpp"

namespace fz::http
{

class client
	: fz::event_handler
	, private logger::modularized
	, private buffer_operator::adder
	, private message_consumer
{
	class performer;

public:
	using cert_verifier = std::function<bool (const tls_session_info &)>;

	struct options: util::options<options, client>
	{
		opt<bool>                  follow_redirects         = o(false);
		opt<uint8_t>               redirects_limits         = o(10);
		opt<std::string>           user_agent               = o(fz::replaced_substrings(PACKAGE_NAME, " ", "-").append("/" PACKAGE_VERSION));
		opt<std::size_t>           max_body_size            = o(10*1024*1024);
		opt<response::handler>     default_response_handler = o();
		opt<headers>               default_request_headers  = o();
		opt<client::cert_verifier> cert_verifier            = o();

		options(){}
	};

	client(thread_pool &pool, event_loop &loop, logger_interface &logger, tls_system_trust_store &trust_store, options opts = {});
	~client() override;

	performer perform(std::string verb, fz::uri uri, std::string body = {}, headers headers = {});

private:
	friend performer;

	void enqueue_request(request &&request, response::handler &&handler);
	void process_queue();
	void close(int err);

private:
	void on_connection_error(int error);
	void on_certificate_verification_event(tls_layer *, tls_session_info &info);
	void on_socket_event(fz::socket_event_source *, fz::socket_event_flag type, int error);
	void on_channel_done_event(channel &, channel::error_type error);
	void operator ()(const event_base &ev) override;

private:
	thread_pool &pool_;
	tls_system_trust_store &trust_store_;
	logger_interface &logger_;
	options opts_;
	std::unique_ptr<securable_socket> socket_;
	channel channel_;

private:
	fz::mutex mutex_;
	std::deque<request> requests_;
	std::deque<response::handler> response_handlers_;

	request current_request_{};
	std::optional<response> current_response_{};
	bool current_response_headers_have_been_parsed_{};
	response::handler current_response_handler_{};

	bool must_secure_socket_{};
	uint8_t redirection_num_{};

private:
	int add_to_buffer() override;
	int process_message_start_line(std::string_view line) override;
	int process_message_header(std::string_view key, std::string_view value) override;
	int process_body_chunk(line_consumer::buffer_string_view chunk) override;
	int process_end_of_message_headers() override;
	int process_end_of_message() override;
};

class client::performer final
{
public:
	~performer();

	void then(response::handler handler);

private:
	friend client;

	performer(client &client, request &&request);

private:
	client &client_;
	request request_;
	response::handler response_handler_;
};


inline client::performer client::perform(std::string verb, fz::uri uri, std::string body, headers headers)
{
	return { *this, { std::move(verb), std::move(uri), std::move(body), std::move(headers) } };
}

inline client::performer::performer(client &client, request &&request)
	: client_(client)
	, request_(std::move(request))
	, response_handler_(client_.opts_.default_response_handler())
{}

inline client::performer::~performer()
{
	client_.enqueue_request(std::move(request_), std::move(response_handler_));
}

inline void client::performer::then(response::handler handler)
{
	response_handler_ = std::move(handler);
}

}

#endif // FZ_HTTP_CLIENT_HPP
