#include <cstdlib>

#include <libfilezilla/local_filesys.hpp>

#include "known_paths.hpp"

#include "util/username.hpp"


#if defined(FZ_WINDOWS)
#	include <libfilezilla/glue/windows.hpp>
#	include <libfilezilla/glue/dll.hpp>
#	include <shlobj.h>
#endif

namespace fz::known_paths {

namespace {

	#if defined (FZ_WINDOWS)
	namespace folderid {

		static constexpr GUID profile       = { 0x5E6C858F, 0x0E22, 0x4760, {0x9A, 0xFE, 0xEA, 0x33, 0x17, 0xB6, 0x71, 0x73} };
		static constexpr GUID local_appdata = { 0xF1B32785, 0x6FBA, 0x4FCF, {0x9D, 0x55, 0x7B, 0x8E, 0x7F, 0x15, 0x70, 0x91} };
		static constexpr GUID program_data  = { 0x62AB5D82, 0xFDC1, 0x4DC3, {0xA9, 0xDD, 0x07, 0x0D, 0x1D, 0x49, 0x5D, 0x97} };
		static constexpr GUID windows       = { 0xF38BF404, 0x1D43, 0x42F2, {0x93, 0x05, 0x67, 0xDE, 0x0B, 0x28, 0xFC, 0x23} };

	}
	#endif

	path_list empty_list() { return {}; }
	path &&same_path(path &&p) { return std::move(p); }
	path_list absolute_filter(native_string_view dirs){
		path_list ret;

		#if defined(FZ_WINDOWS)
			auto sep = fzT(";");
		#else
			auto sep = fzT(":");
		#endif

		for (auto e: strtok_view(dirs, sep)) {
			if (path p{e}; p.is_absolute())
				ret.push_back(std::move(p));
		}
		return ret;
	}

	template <typename Var, typename IfFunc, typename ElseFunc>
	auto get(Var && var, IfFunc if_func, ElseFunc else_func) -> std::decay_t<decltype (if_func(fz::native_string()))>
	{
		native_string value;

		#if defined (FZ_WINDOWS)
			static auto get = reinterpret_cast<decltype(&SHGetKnownFolderPath)>(shdlls::get().shell32_["SHGetKnownFolderPath"]);
			static auto free = reinterpret_cast<decltype(&CoTaskMemFree)>(shdlls::get().ole32_["CoTaskMemFree"]);

			if (get && free) {
				wchar_t *wret{};

				if (SUCCEEDED(get(var, 0, nullptr, &wret)))
					value = native_string(wret);

				free(wret);
			}
		#else
			auto v = getenv(var);
			if (v && v[0])
				value = fz::to_native(v);
		#endif

		if (!value.empty())
			return if_func(std::move(value));

		return else_func();
	}

	#if defined(FZ_WINDOWS)
	#	define var(Unix, Windows) Windows
	#elif defined(FZ_UNIX) || defined(FZ_MAC)
	#	define var(Unix, Windows) Unix
	#else
	#   error "Your platform is not supported yet."
	#endif

	const path &HOME()  {
		static const path p = get(var("HOME", folderid::profile), same_path, []{ return native_string(1, local_filesys::path_separator); });
		return p;
	}
}

const path &data::home() {
	static const path p = get(var("XDG_DATA_HOME", folderid::local_appdata), same_path, []{ return HOME() / fzT(".local") / fzT("share"); });
	return p;
}

const path_list &data::dirs() {
	static const path_list pl = get(var("XDG_DATA_DIRS", folderid::local_appdata), absolute_filter, empty_list);
	return pl;
}

const path &config::home() {
	static const path p = get(var("XDG_CONFIG_HOME", folderid::local_appdata), same_path, []{ return HOME() / fzT(".config"); });
	return p;
}

const path_list &config::dirs() {
	static const path_list pl = get(var("XDG_CONFIG_DIRS", folderid::local_appdata), absolute_filter, empty_list);
	return pl;
}

const path &config::effective_home()
{
#if defined(FZ_WINDOWS)
	bool is_system = fz::stricmp(util::get_current_user_name(), L"SYSTEM") == 0;

	if (is_system)
		return windows::program_data();

#endif
	return home();
}

#if defined(FZ_WINDOWS)
namespace windows {
	const path &program_data()
	{
		static const path p = get(folderid::program_data, same_path, config::home);
		return p;
	}

	const path &windir()
	{
		static const path p = get(folderid::windows, same_path, [] { return L"C:\\Windows"; });
		return p;
	}
}
#endif

}
