#include <libfilezilla/format.hpp>

#include "../logger/file.hpp"

#include "../util/filesystem.hpp"
#include "../util/parser.hpp"

namespace fz::logger {

file::file(file::options opts)
	: stdio(stderr)
{
	set_options(std::move(opts));
}

void file::set_options(const file::options &opts)
{
	bool emit_start_line = opts.start_line() && opts.name() != opts_.name();
	set_all(opts.enabled_types());

	{
		auto lock = buffer_.lock();
		opts_ = opts;

		open(fz::file::creation_flags::existing);
	}

	if (emit_start_line)
		log(logmsg::status, L"===== New logging started =====");
}

void file::do_log(logmsg::type t, std::wstring &&msg)
{
	auto buffer = buffer_.lock();

	auto now = format_message(*buffer, t, std::move(msg), opts_.include_headers(), opts_.remove_cntrl(), opts_.split_lines());
	if (!file_.opened()) {
		stdio::log_formatted_message(*buffer);
		return;
	}

	maybe_rotate(now);

	while (buffer->size()) {
		auto to_consume = file_.write(buffer->get(), (int64_t)buffer->size());

		if (to_consume < 0) {
			stdio::log_formatted_message(*buffer);
			break;
		}

		file_size_ += to_consume;

		buffer->consume((std::size_t)to_consume);
	}
}

void file::open(fz::file::creation_flags flags)
{
	file_.close();
	file_dt_ = {};

	if (!opts_.name().empty() && opts_.max_file_size() > 0) {
		file_.open(opts_.name(), fz::file::mode::writing, flags | fz::file::creation_flags::current_user_and_admins_only);

		if (flags & fz::file::creation_flags::existing)
			file_.seek(0, fz::file::seek_mode::end);

		auto size = file_.size();
		file_size_ = size < 0 ? 0 : size;
		file_dt_ = local_filesys::get_modification_time(opts_.name());
	}
}

void file::maybe_rotate(const datetime &now)
{
	// Check if rotation criteria are satisfied
	switch (opts_.rotation_type()) {
		case options::size_based:
			if (file_size_ < opts_.max_file_size())
				return;
			break;

		case options::daily:
			static const auto origin = datetime(0, datetime::days);
			if ((now - origin).get_days() == (file_dt_ - origin).get_days())
				return;
			break;
	}

	// In which case, rotate.

	file_.close();
	file_size_ = 0;

	if (!opts_.name().empty()) {
		std::decay_t<decltype(opts_.max_amount_of_rotated_files())> max = 0;

		// Search for the rotated file with the highest number
		for (const auto &file: util::fs::native_path(opts_.name()).parent()) {
			util::parseable_range r(file.str());
			decltype(max) cur = 0;

			if (match_string(r, opts_.name()) && lit(r, '.') && parse_int(r, cur) && eol(r)) {
				// The name matches the pattern <log_file_name>.<number>
				if (max < cur)
					max = cur;
			}
		}

		if (max < opts_.max_amount_of_rotated_files())
			max += 1;
		else
			max = opts_.max_amount_of_rotated_files();

		for (auto i = max; i-- > 0;) {
			const auto &from = (i == 0)
				? opts_.name()
				: sprintf(fzT("%s.%d"), opts_.name(), i);

			const auto &to = sprintf(fzT("%s.%d"), opts_.name(), i+1);

			// Silently fail if the file's not there yet.
			fz::rename_file(from, to, false);
		}

		open(fz::file::creation_flags::empty);
	}
}

}
