#include <libfilezilla/glue/wx.hpp>

#include <set>
#include <tuple>

#include <wx/grid.h>
#include <wx/sizer.h>
#include <wx/button.h>
#include <wx/log.h>

#include "fluidcolumnlayoutmanager.hpp"
#include "addressinfolisteditor.hpp"
#include "locale.hpp"
#include "glue.hpp"
#include "helpers.hpp"
#include "gridcellenumrendererex.hpp"

#include "../filezilla/hostaddress.hpp"

enum: int {
	address_col = 0,
	port_col = 1,
	proto_col = 2
};

class AddressInfoListEditor::Grid: public wxGrid {
public:
	Grid() = default;
	bool Create(wxWindow *parent);
};


class AddressInfoListEditor::Table: public wxGridTableBase {
public:
	Table(bool edit_proto);

	void SetLists(std::initializer_list<ListInfo> lists);
	void AddRow();
	void RemoveSelectedRows();
	int GetNumberRows() override;
	bool Validate();
	bool CanGetValueAs(int, int col, const wxString &name) override;
	bool CanSetValueAs(int, int col, const wxString &name) override;
	long GetValueAsLong(int row, int col) override;
	void SetValueAsLong(int row, int col, long value) override;

	wxString GetRowLabelValue(int row) override;

	int GetNumberLists();

	bool TransferDataFromWindow();

private:
	friend AddressInfoListEditor;

	void SetAddressAndPortValidator(validator_t validator);
	wxString GetMatchingAddress(const wxString &address, unsigned int port, bool any_is_equivalent);

	int GetNumberCols() override;
	wxString GetValue(int row, int col) override;
	void SetValue(int row, int col, const wxString &value) override;
	wxString GetColLabelValue(int col) override;

	void SetView(wxGrid *) override;

private:
	void UpdateProtocols();

	std::vector<std::vector<fz::tcp::address_info> *> lists_;
	std::vector<std::pair<fz::tcp::address_info::tls_mode_t, std::vector<fz::tcp::address_info> *>> proto2list_;
	std::vector<std::pair<fz::tcp::address_info::tls_mode_t, std::vector<fz::tcp::address_info> *>> unknown_proto2list_;
	wxString proto_names_;

	struct Line
	{
		wxString address{};
		unsigned int port{};
		long proto{};
	};

	std::vector<Line> lines_;

	bool edit_proto_{};
	bool modified_{};

	wxGridCellAttr *unknown_attr_{};
	validator_t validator_;
};

bool AddressInfoListEditor::Grid::Create(wxWindow *parent)
{
	if (!wxGrid::Create(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxWANTS_CHARS | wxBORDER_THEME))
		return false;

	UseNativeColHeader();
	Bind(wxEVT_GRID_LABEL_RIGHT_CLICK, [](auto &){});

	SetDefaultCellAlignment(wxALIGN_LEFT, wxALIGN_BOTTOM);
	SetRowLabelSize(0);
	AutoSize();
	DisableDragRowSize();
	DisableDragColMove();
	SetColLabelAlignment(wxALIGN_LEFT, wxALIGN_CENTER);

	Bind(wxEVT_CHAR_HOOK, [&](wxKeyEvent &ev) {
	   if (ev.GetKeyCode() == WXK_ESCAPE && IsCellEditControlEnabled()) {
		   auto editor = GetCellEditor(GetGridCursorRow(), GetGridCursorCol());
		   editor->Reset();
		   editor->DecRef();
		   DisableCellEditControl();
	   }
	   else
		   ev.Skip();
	});

	return true;
}

bool AddressInfoListEditor::Create(wxWindow *parent, bool edit_proto)
{
	if (!wxPanel::Create(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL | wxNO_BORDER, wxS("AddressInfoListEditor")))
		return false;

	wxHBox(this, 0) = {
		{ wxSizerFlags(1).Expand(), grid_ = wxCreate<Grid>(this) },
		{ wxSizerFlags(0).Expand(), wxVBox(this, 0) = {
			{ 0, add_button_ = new wxButton(this, wxID_ANY, _S("&Add")) },
			{ 0, remove_button_ = new wxButton(this, wxID_ANY, _S("Re&move")) }
		}}
	};

	if (table_ = new Table(edit_proto); !grid_->SetTable(table_, true, wxGrid::wxGridSelectRows))
		return false;

	add_button_->Bind(wxEVT_BUTTON, [this](auto){
		grid_->SetFocus();
		table_->AddRow();

		remove_button_->Enable();

		GetParent()->Layout();
   });

	remove_button_->Bind(wxEVT_BUTTON, [this](auto){
		table_->RemoveSelectedRows();
		if (table_->GetNumberRows() == 0)
			remove_button_->Disable();

		GetParent()->Layout();
	});

	SetLists({});

	return true;
}

void AddressInfoListEditor::SetLists(std::initializer_list<ListInfo> lists)
{
	table_->SetLists(lists);

	if (table_->GetNumberLists() > 0) {
		add_button_->Enable();

		grid_->Enable();

		if (table_->GetNumberRows() > 0)
			remove_button_->Enable();
	}
	else {
		add_button_->Disable();
		remove_button_->Disable();
		grid_->Disable();
	}
}

void AddressInfoListEditor::SetAddressAndPortValidator(validator_t validator)
{
	table_->SetAddressAndPortValidator(std::move(validator));
}

wxString AddressInfoListEditor::GetMatchingAddress(const wxString &address, unsigned int port, bool any_is_equivalent)
{
	return table_->GetMatchingAddress(address, port, any_is_equivalent);
}

bool AddressInfoListEditor::Validate()
{
	if (!wxPanel::Validate())
		return false;

	if (!grid_->IsEnabled())
		return true;

	grid_->DisableCellEditControl();
	return table_->Validate();
}

bool AddressInfoListEditor::TransferDataFromWindow()
{
	if (!wxPanel::TransferDataFromWindow())
		return false;

	if (!grid_->IsEnabled())
		return true;

	grid_->DisableCellEditControl();
	return table_->TransferDataFromWindow();
}

void AddressInfoListEditor::Table::SetView(wxGrid *grid)
{
	auto cl = new wxFluidColumnLayoutManager(grid);
	cl->SetColumnWeight(address_col, 1);
	cl->SetColumnWeight(port_col, 1);

	if (edit_proto_)
		cl->SetColumnWeight(proto_col, 3);

	wxGridTableBase::SetView(grid);
	UpdateProtocols();
}

void AddressInfoListEditor::Table::UpdateProtocols()
{
	wxASSERT(GetView() != nullptr);

	static const wxString PROTO_NAMES = wxT("proto_names");

	GetView()->RegisterDataType(PROTO_NAMES, new wxGridCellEnumRendererEx(proto_names_), new wxGridCellEnumEditor(proto_names_));
	GetView()->SetColFormatCustom(proto_col, PROTO_NAMES);
}


bool AddressInfoListEditor::Table::Validate()
{
	static const auto address_is_invalid = _S("Address is invalid");
	static const auto port_out_of_range = _S("Port number must be between 1 and 65535");
	static const auto address_and_port_must_be_unique = [](const Line &l) {
		return _F("Duplicated address %s", fz::join_host_and_port(l.address.ToStdWstring(), l.port));
	};

	auto invalid = [&](const wxString &msg, int row, int col) {
		if (col >= 0)
			wxMsgError(_F("At row %d, column %d.\n\n%s.", row+1, col+1, msg));
		else
			wxMsgError(_F("At row %d.\n\n%s.", row+1, msg));

		GetView()->CallAfter([grid = GetView(), row, col] {
			grid->SetFocus();
			grid->SelectRow(row);

			if (col >= 0)
				grid->GoToCell(row, col);

			if (grid->CanEnableCellControl())
				grid->EnableCellEditControl();
		});

		return false;
	};

	auto already_seen = [&, set = std::set<std::tuple<wxString, unsigned int>>{}](const Line &i) mutable {
		auto [it, inserted] = set.insert({i.address, i.port});
		return !inserted;
	};

	int row = 0;
	for (const auto &i: lines_) {
		if (i.port < 1 || i.port > 65535)
			return invalid(port_out_of_range, row, port_col);

		if (validator_)
		if (const wxString &error = validator_(i.address, i.port); !error.empty())
			return invalid(error, row, -1);

		static auto check_ip = [](const auto &str, auto &ip) {
			fz::util::parseable_range r(str);
			return parse_ip(r, ip) && eol(r);
		};

		if (fz::hostaddress::ipv4_host ip; check_ip(i.address.ToStdWstring(), ip));
		else
		if (fz::hostaddress::ipv6_host ip; check_ip(i.address.ToStdWstring(), ip));
		else
			return invalid(address_is_invalid, row, 0);

		if (already_seen(i))
			return invalid(address_and_port_must_be_unique(i), row, -1);

		++row;
	}

	return true;
}

wxString AddressInfoListEditor::Table::GetRowLabelValue(int)
{
	return wxEmptyString;
}

int AddressInfoListEditor::Table::GetNumberLists()
{
	return int(lists_.size());
}

int AddressInfoListEditor::Table::GetNumberRows()
{
	return int(lines_.size());
}


int AddressInfoListEditor::Table::GetNumberCols()
{
	return 2 + edit_proto_;
}

wxString AddressInfoListEditor::Table::GetValue(int row, int col) {
	if (row < 0 || col < 0)
		return wxEmptyString;

	wxCHECK_MSG(0 <= row && row < (int)GetNumberRows(), wxEmptyString, "Not enough rows");
	wxCHECK_MSG(0 <= col && col < GetNumberCols(), wxEmptyString, "Not enough cols");

	auto &line = lines_[(size_t)row];

	switch (col) {
		case address_col: return line.address;
		case port_col: return wxString::Format("%u", line.port);
		case proto_col:
			if (line.proto < 0)
				return _S("Unknown");
			else
				return wxString::Format("%ld", line.proto);
	}

	return wxEmptyString;
}

bool AddressInfoListEditor::Table::CanGetValueAs(int row, int col, const wxString &name)
{
	wxCHECK_MSG(0 <= row && row < GetNumberRows(), 0, "Not enough rows");
	wxCHECK_MSG(0 <= col && col < GetNumberCols(), 0, "Not enough cols");

	if (col == proto_col && name == wxGRID_VALUE_NUMBER) {
		if (lines_[std::size_t(row)].proto < 0)
			return false;

		return true;
	}


	return name == wxGRID_VALUE_STRING;
}

bool AddressInfoListEditor::Table::CanSetValueAs(int row, int col, const wxString &name)
{
	wxCHECK_MSG(0 <= row && row < GetNumberRows(), 0, "Not enough rows");
	wxCHECK_MSG(0 <= col && col < GetNumberCols(), 0, "Not enough cols");

	if (col == proto_col && name == wxGRID_VALUE_NUMBER)
		return true;

	return name == wxGRID_VALUE_STRING;
}

long AddressInfoListEditor::Table::GetValueAsLong(int row, int col)
{
	if (row < 0 || col < 0)
		return 0;

	wxCHECK_MSG(0 <= row && row < GetNumberRows(), 0, "Not enough rows");
	wxCHECK_MSG(0 <= col && col < GetNumberCols(), 0, "Not enough cols");

	if (col == proto_col)
		return 	lines_[(size_t)row].proto;

	return 0;
}

void AddressInfoListEditor::Table::SetValueAsLong(int row, int col, long value)
{
	if (row < 0 || col < 0)
		return;

	wxCHECK_RET(0 <= row && row < GetNumberRows(), "Not enough rows");
	wxCHECK_RET(0 <= col && col < GetNumberCols(), "Not enough cols");

	if (col == proto_col) {
		if (auto &p = lines_[(size_t)row].proto; p != value) {
			modified_ = true;

			p = value;
		}
	}
}

wxString AddressInfoListEditor::Table::GetMatchingAddress(const wxString &address, unsigned int port, bool any_is_equivalent)
{
	auto it = std::find_if(lines_.begin(), lines_.end(), [&](const Line &l) {
		auto h1 = fz::hostaddress(l.address.ToStdWstring(), fz::hostaddress::format::ipvx, l.port);
		auto h2 = fz::hostaddress(address.ToStdWstring(), fz::hostaddress::format::ipvx, port);

		return h1.equivalent_to(h2, any_is_equivalent) && l.port == port;
	});

	if (it != lines_.end())
		return it->address;

	return {};
}


void AddressInfoListEditor::Table::SetValue(int row, int col, const wxString &value) {
	if (row < 0 || col < 0)
		return;

	wxCHECK_RET(0 <= row && row < GetNumberRows(), "Not enough rows");
	wxCHECK_RET(0 <= col && col < GetNumberCols(), "Not enough cols");

	auto &line = lines_[(std::size_t)row];

	switch (col) {
		case address_col:
			if (auto &a = line.address; a != value) {
				modified_ = true;
				a = value;
			}

			break;

		case port_col: {
			auto np = fz::to_integral(value.ToStdWstring(), 0u);
			if (auto &p = line.port; p != np) {
				modified_ = true;
				p = np;
			}

			break;
		}

		case proto_col: {
			auto np = fz::to_integral(value.ToStdWstring(), 0l);

			if (auto &p = line.proto; p != np) {
				modified_ = true;
				p = np;
			}

			break;
		}
	}
}


wxString AddressInfoListEditor::Table::GetColLabelValue(int col)
{
	wxCHECK_MSG( (col >= 0 && col < GetNumberCols()),
				wxEmptyString,
				wxS("invalid column index in MountTable::Table") );

	switch (col) {
		case address_col: return _S("Address");
		case port_col: return _S("Port");
		case proto_col: return _S("Protocol");
	}

	return wxEmptyString;
}


AddressInfoListEditor::Table::Table(bool edit_proto)
	: edit_proto_(edit_proto)
{
}

void AddressInfoListEditor::Table::SetLists(std::initializer_list<ListInfo> lists) {
	auto old_rows = GetNumberRows();

	lists_.clear();
	proto_names_.clear();
	proto2list_.clear();
	lines_.clear();
	modified_ = false;

	for (auto &i: lists) {
		// If either the list or the tls map is not provided, ignore this item.
		if (!i.list || i.tls2name_map.size() == 0)
			continue;

		// Build the mapping
		lists_.push_back(i.list);

		auto tls_mode2proto = [&, first_n = long(proto2list_.size())](fz::tcp::address_info::tls_mode_t tls_mode) {
			auto n = first_n;

			for (auto &tn: i.tls2name_map) {
				if (tls_mode == tn.first)
					return n;

				n += 1;
			}

			// Lines with tls modes not mapped to anything are still displayed, but not editable.
			unknown_proto2list_.emplace_back(tls_mode, i.list);
			return -long(unknown_proto2list_.size());
		};

		for (auto &tn: i.tls2name_map) {
			if (proto_names_.size() > 0)
				proto_names_ += wxT(",");

			proto_names_ += tn.second;
			proto2list_.push_back({tn.first, i.list});
		}

		// Populate the internal, merged, list
		for (auto &l: *i.list)
			lines_.push_back({fz::to_wxString(l.address), l.port, tls_mode2proto(l.tls_mode)});
	}

	UpdateProtocols();

	wxGridUpdateLocker updater(GetView());

	if (auto delta_rows = GetNumberRows()-old_rows; delta_rows > 0)	{
		wxGridTableMessage msg(this, wxGRIDTABLE_NOTIFY_ROWS_APPENDED, delta_rows);
		GetView()->ProcessTableMessage(msg);
	}
	else
	if (delta_rows < 0) {
		wxGridTableMessage msg(this, wxGRIDTABLE_NOTIFY_ROWS_DELETED, old_rows+delta_rows, -delta_rows);
		GetView()->ProcessTableMessage(msg);
	}

	for (int r = 0; size_t(r) < lines_.size(); ++r) {
		if (auto &l = lines_[size_t(r)]; l.proto < 0) {
			if (!unknown_attr_) {
				unknown_attr_ = new wxGridCellAttr;
				unknown_attr_->SetReadOnly();
				unknown_attr_->SetTextColour(*wxLIGHT_GREY);
				unknown_attr_->SetFont(GetView()->GetFont().MakeItalic());
			}
			else
				unknown_attr_->IncRef();

			GetView()->SetRowAttr(r, unknown_attr_);
		}
	}

	wxGridTableMessage msg(this, wxGRIDTABLE_REQUEST_VIEW_GET_VALUES);
	GetView()->ProcessTableMessage(msg);
}

bool AddressInfoListEditor::Table::TransferDataFromWindow()
{
	if (!modified_)
		return true;

	for (auto &l: lists_) {
		wxASSERT(l != nullptr);
		l->clear();
	}

	for (auto &l: lines_) {
		fz::tcp::address_info ::tls_mode_t tls_mode{};
		std::vector<fz::tcp::address_info> *list{};

		if (l.proto < 0) {
			auto protoid = std::size_t(-1l - l.proto);
			wxCHECK_MSG(protoid < unknown_proto2list_.size(), false, "Invalid protocol id");

			std::tie(tls_mode, list) = unknown_proto2list_[protoid];
		}
		else {
			auto protoid = std::size_t(l.proto);
			wxCHECK_MSG(protoid < proto2list_.size(), false, "Invalid protocol id");

			std::tie(tls_mode, list) = proto2list_[protoid];
		}

		wxASSERT(list);

		list->push_back({fz::to_utf8(l.address), l.port, tls_mode});
	}

	modified_ = false;
	return true;
}

void AddressInfoListEditor::Table::SetAddressAndPortValidator(validator_t validator)
{
	validator_ = std::move(validator);
}

void AddressInfoListEditor::Table::AddRow()
{
	wxASSERT(proto2list_.size() > 0);

	wxGridUpdateLocker updater(GetView());
	modified_ = true;

	lines_.push_back({});

	wxGridTableMessage msg(this, wxGRIDTABLE_NOTIFY_ROWS_APPENDED, 1);
	GetView()->ProcessTableMessage(msg);

	GetView()->CallAfter([this]{
		GetView()->SelectRow(int(lines_.size()-1));
		GetView()->GoToCell(int(lines_.size()-1), 0);
		GetView()->EnableCellEditControl();
		GetView()->SetFocus();
	});
}

void AddressInfoListEditor::Table::RemoveSelectedRows()
{
	wxGridUpdateLocker updater(GetView());
	modified_ = true;

	for (auto &row: GetView()->GetSelectedRows()) {
		lines_.erase(std::next(lines_.begin(), row));
		wxGridTableMessage msg(this, wxGRIDTABLE_NOTIFY_ROWS_DELETED, row, 1);
		GetView()->ProcessTableMessage(msg);
	}
}
