#ifndef HELPERS_HPP
#define HELPERS_HPP

#include <variant>
#include <functional>
#include <stack>
#include <queue>

#include <libfilezilla/logger.hpp>

#include <wx/sizer.h>
#include <wx/stattext.h>
#include <wx/notebook.h>
#include <wx/treebook.h>
#include <wx/choicebk.h>
#include <wx/panel.h>
#include <wx/msgdlg.h>
#include <wx/wizard.h>
#include <wx/hyperlink.h>
#include <wx/textctrl.h>
#include <wx/simplebook.h>
#include <wx/timer.h>
#include <wx/checkbox.h>

#include "dialogex.hpp"
#include "locale.hpp"
#include "wrappedtext.hpp"

class wxTitleCtrl: public wxPanel
{
public:
	wxTitleCtrl(wxWindow *parent, const wxString &title, int style = 0);

	wxStaticText *GetLabelCtrl();

private:
	wxStaticText *label_;
};

wxTitleCtrl *wxTitle(wxWindow *parent, const wxString &title, int style = 0);

template <typename Sizer, typename Object, typename Derived>
struct wxSizerWrapper;

struct wxSizerProxy
{
	wxSizerProxy(wxWindow *parent, wxSizer *sizer);
	wxSizerProxy(wxSizerProxy &&rhs) noexcept;

	wxSizerProxy(const wxSizerProxy &) = delete;
	wxSizerProxy &operator=(const wxSizerProxy &) = delete;
	wxSizerProxy &operator=(wxSizerProxy &&) = delete;

	~wxSizerProxy();

	operator wxSizer *();
	wxSizer *operator->();

private:
	wxWindow *parent_;
	wxSizer *sizer_;
};

int wxDlg2Px(wxWindow *w, int dlg);

void wxValidate(wxWindow *w, std::function<bool()> func);
void wxTransferDataToWindow(wxWindow *w, std::function<bool()> func);
void wxTransferDataFromWindow(wxWindow *w, std::function<bool()> func);

bool wxCopyToClipboard(const wxString &str, bool is_html = false);

wxStaticText *wxLabel(wxWindow *parent, const wxString &text, int style = 0);
wxStaticText *wxEscapedLabel(wxWindow *parent, const wxString &text, int style = 0);

wxWrappedText *wxWText(wxWindow *parent, const wxString &text);

void wxSetWindowLogger(wxWindow *win, fz::logger_interface *logger);
fz::logger_interface *wxGetWindowLogger(wxWindow *win);

void wxMsg(int flags, const wxString &title, const wxString &msg, const wxString &ext);

template <typename... Args>
void wxMsgError(const wxString &msg, const Args&... args)
{
	wxMsg(wxOK | wxCENTRE | wxICON_ERROR, _S("Error"), fz::to_wxString(fz::sprintf(msg.ToStdWstring(), args...)), wxEmptyString);
}

template <typename... Args>
void wxMsgErrorExt(const wxString &msg, const wxString &ext, const Args &... args)
{
	wxMsg(wxOK | wxCENTRE | wxICON_ERROR, _S("Error"), fz::to_wxString(fz::sprintf(msg.ToStdWstring(), args...)), fz::to_wxString(fz::sprintf(ext.ToStdWstring(), args...)));
}

template <typename... Args>
void wxMsgSuccess(const wxString &msg, const Args &... args)
{
	wxMsg(wxOK | wxCENTRE | wxICON_INFORMATION, _S("Success"), fz::to_wxString(fz::sprintf(msg.ToStdWstring(), args...)), wxEmptyString);
}

template <typename... Args>
void wxMsgSuccessExt(const wxString &msg, const wxString &ext, const Args &... args)
{
	wxMsg(wxOK | wxCENTRE | wxICON_INFORMATION, _S("Success"), fz::to_wxString(fz::sprintf(msg.ToStdWstring(), args...)), fz::to_wxString(fz::sprintf(ext.ToStdWstring(), args...)));
}

template <typename... Args>
void wxMsgWarning(const wxString &msg, const Args &... args)
{
	wxMsg(wxOK | wxCENTRE | wxICON_EXCLAMATION, _S("Warning"), fz::to_wxString(fz::sprintf(msg.ToStdWstring(), args...)), wxEmptyString);
}

template <typename... Args>
void wxMsgWarningExt(const wxString &msg, const wxString &ext, const Args &... args)
{
	wxMsg(wxOK | wxCENTRE | wxICON_EXCLAMATION, _S("Warning"), fz::to_wxString(fz::sprintf(msg.ToStdWstring(), args...)), fz::to_wxString(fz::sprintf(ext.ToStdWstring(), args...)));
}

template <typename SizerObject>
class wxSizerPair: public std::pair<wxSizerFlags, SizerObject>
{
public:
	using std::pair<wxSizerFlags, SizerObject>::pair;
	using std::pair<wxSizerFlags, SizerObject>::operator=;

	template <typename T, std::enable_if_t<std::is_convertible_v<T, SizerObject>>* = nullptr>
	wxSizerPair(T obj)
		: wxSizerPair{wxSizerFlags(0), std::move(obj)}
	{}
};

template <typename T>
std::pair<wxSizerFlags, T> operator >>=(wxSizerFlags f, T t)
{
	return { std::move(f), std::move(t) };
}

using wxBoxSizerObject = std::variant<wxWindow *, wxSizer *, int>;
void wxAddToSizer(wxBoxSizer *sizer, wxWindow *parent, std::initializer_list<wxSizerPair<wxBoxSizerObject>> elems, int gap);

using wxGridSizerObject = wxBoxSizerObject; //std::variant<wxWindow *, wxSizer *, wxString>;
void wxAddToSizer(wxFlexGridSizer *sizer, wxWindow *parent, std::initializer_list<wxSizerPair<wxGridSizerObject>> elems, wxAlignment default_alignment = wxALIGN_CENTER_VERTICAL);

template <typename Sizer, typename Object, typename Derived>
struct wxSizerWrapper
{
	wxSizerWrapper(wxWindow *parent, Sizer *sizer)
		: parent_(parent)
		, sizer_(sizer)
	{}

	wxSizerProxy operator =(std::initializer_list<wxSizerPair<Object>> elems) &&
	{
		static_cast<Derived &>(*this).wxAddToSizer(sizer_, parent_, std::move(elems));

		return {parent_, sizer_};
	}

	wxSizerProxy operator =(std::pair<wxSizerFlags, Object> elem) &&
	{
		return (std::move(static_cast<Derived &>(*this))) = { wxSizerPair<Object>(std::move(elem.first), std::move(elem.second)) };
	}

	wxSizerProxy operator =(Object elem) &&
	{
		return (std::move(static_cast<Derived &>(*this))) = { wxSizerPair<Object>(wxSizerFlags(1), std::move(elem)) };
	}

protected:
	void wxAddToSizer(Sizer *sizer, wxWindow *parent, std::initializer_list<wxSizerPair<Object>> elems)
	{
		::wxAddToSizer(sizer, parent, elems);
	}

	wxWindow *const parent_;
	Sizer *sizer_;
};

struct wxGridSizerWrapper: wxSizerWrapper<wxFlexGridSizer, wxGridSizerObject, wxGridSizerWrapper>
{
	using wxSizerWrapper::wxSizerWrapper;
	using wxSizerWrapper::operator=;

	wxGridSizerWrapper(wxWindow *parent, wxFlexGridSizer *sizer, wxAlignment default_alignment = wxALIGN_CENTER_VERTICAL);

	void wxAddToSizer(wxFlexGridSizer *sizer, wxWindow *parent, std::initializer_list<wxSizerPair<wxGridSizerObject>> elems);

private:
	wxAlignment default_alignment_;
};

enum: int { wxDefaultPadding = 2 };
enum: int { wxDefaultGap = 2 };

struct wxPadding
{
	constexpr wxPadding(wxDirection dir, int pad = wxDefaultPadding, int offset = 0)
		: dir(dir)
		, pad(pad)
		, offset(offset)
	{}

	constexpr wxPadding(int pad = wxDefaultPadding)
		: wxPadding(wxALL, pad)
	{}

	wxDirection dir;
	int pad;
	int offset;
};

struct wxBoxSizerWrapper: wxSizerWrapper<wxBoxSizer, wxBoxSizerObject, wxBoxSizerWrapper>
{
	wxBoxSizerWrapper(wxWindow *parent, wxBoxSizer *sizer, wxPadding padding = wxDefaultPadding, int gap = wxDefaultGap);

	wxSizerProxy operator =(std::initializer_list<wxSizerPair<wxBoxSizerObject>> elems) &&;

	using wxSizerWrapper::operator=;

	void wxAddToSizer(wxBoxSizer *sizer, wxWindow *parent, std::initializer_list<wxSizerPair<wxBoxSizerObject>> elems);

protected:
	int padding_;
	wxDirection padding_direction_;
	int gap_;
};

wxBoxSizerWrapper wxHBox(wxWindow *parent, wxPadding padding = wxDefaultPadding, int gap = wxDefaultGap);
wxBoxSizerWrapper wxVBox(wxWindow *parent, wxPadding padding = wxDefaultPadding, int gap = wxDefaultGap);
wxBoxSizerWrapper wxStaticHBox(wxWindow *p, const wxString &label = wxEmptyString, wxPadding padding = wxDefaultPadding, int gap = wxDefaultGap);
wxBoxSizerWrapper wxStaticVBox(wxWindow *p, const wxString &label = wxEmptyString, wxPadding padding = wxDefaultPadding, int gap = wxDefaultGap);

constexpr inline auto wxEmptySpace = std::pair{1, -1};

struct wxGBoxGrowableParams
{
	wxGBoxGrowableParams(std::size_t idx, int proportion = 0)
		: idx(idx)
		, proportion(proportion)
	{}

	const std::size_t idx;
	const int proportion;
};

inline static const wxSize wxGBoxDefaultGap = {wxDefaultGap, wxDefaultGap};
wxGridSizerWrapper wxGBox(wxWindow *parent, int cols, std::initializer_list<wxGBoxGrowableParams> growable_cols = {}, std::initializer_list<wxGBoxGrowableParams> growable_rows = {}, wxSize gap = wxGBoxDefaultGap, wxAlignment default_alignment = wxALIGN_CENTER_VERTICAL);

struct wxValidateOnlyIfCurrentPageTag
{
	virtual ~wxValidateOnlyIfCurrentPageTag() = default;
};

template <typename W = wxPanel, std::enable_if_t<std::is_base_of_v<wxWindow,W>>* = nullptr>
struct wxValidateOnlyIfCurrentPage: W, virtual wxValidateOnlyIfCurrentPageTag
{
	using W::W;

	bool Validate() override
	{
		if (auto b = dynamic_cast<wxBookCtrlBase *>(this->GetParent())) {
			return b->GetCurrentPage() != this || W::Validate();
		}

		return W::Validate();
	}

	bool TransferDataToWindow() override
	{
		if (auto b = dynamic_cast<wxBookCtrlBase *>(this->GetParent())) {
			return b->GetCurrentPage() != this || W::TransferDataToWindow();
		}

		return W::TransferDataToWindow();
	}

	bool TransferDataFromWindow() override
	{
		if (auto b = dynamic_cast<wxBookCtrlBase *>(this->GetParent())) {
			return b->GetCurrentPage() != this || W::TransferDataFromWindow();
		}

		return W::TransferDataFromWindow();
	}
};

struct wxCreator
{
	template <typename T, typename SFINAE = void>
	struct BookCtrlFix: T
	{
		using T::T;
	};

	template <typename T>
	struct BookCtrlFix<T, std::enable_if_t<std::is_base_of_v<wxBookCtrlBase, T>>>: T
	{
		using T::T;

		// BUG: see https://trac.wxwidgets.org/ticket/16878
		bool HasTransparentBackground() override
		{
			return true;
		}

		bool Validate() override
		{
			if constexpr (!std::is_base_of_v<wxValidateOnlyIfCurrentPageTag, T>) {
				bool not_overridden = typeid(&wxWindow::Validate) == typeid(&T::Validate);
				wxASSERT_MSG(not_overridden, "The book control you're using has overridden the Validate method: this is not supported");

				if (not_overridden)
					return apply_to_all_pages(this, &wxWindow::Validate, "Validate");
			}

			return T::Validate();
		}

		bool TransferDataFromWindow() override
		{
			if constexpr (!std::is_base_of_v<wxValidateOnlyIfCurrentPageTag, T>) {
				bool not_overridden = typeid(&wxWindow::TransferDataFromWindow) == typeid(&T::TransferDataFromWindow);
				wxASSERT_MSG(not_overridden, "The book control you're using has overridden the TransferDataFromWindow method: this is not supported");

				if (not_overridden)
					return apply_to_all_pages(this, &wxWindow::TransferDataFromWindow, "TransferDataFromWindow");
			}

			return T::TransferDataFromWindow();
		}

		bool TransferDataToWindow() override
		{
			if constexpr (!std::is_base_of_v<wxValidateOnlyIfCurrentPageTag, T>) {
				bool not_overridden = typeid(&wxWindow::TransferDataToWindow) == typeid(&T::TransferDataToWindow);
				wxASSERT_MSG(not_overridden, "The book control you're using has overridden the TransferDataToWindow method: this is not supported");

				if (not_overridden)
					return apply_to_all_pages(this, &wxWindow::TransferDataToWindow, "TransferDataToWindow");
			}

			return T::TransferDataToWindow();
		}
	};

	template <typename W, typename Args, typename = void>
	struct supports_two_phases_creation: std::false_type {};

	template <typename W, typename... Args>
	struct supports_two_phases_creation<W, std::tuple<Args...>, std::void_t<decltype(W(), std::declval<W>().Create(std::declval<Args>()...) == true)>>: std::true_type{};

	template <typename W, typename... Args>
	std::enable_if_t<
		supports_two_phases_creation<W, std::tuple<Args...>>::value,
	std::add_pointer_t<W>>
	static create(Args &&... args)
	{
		W* w = new BookCtrlFix<W>();

		if (!w->Create(std::forward<Args>(args)...)) {
			abort();
		}

		return w;
	}

	template <typename W, typename... Args>
	std::enable_if_t<
		std::is_constructible_v<W, Args...> &&
		!wxCreator::supports_two_phases_creation<W, std::tuple<Args...>>::value,
	std::add_pointer_t<W>>
	static create(Args &&... args)
	{
		return new BookCtrlFix<W>(std::forward<Args>(args)...);
	}

private:
	static bool apply_to_all_pages(wxBookCtrlBase *b, bool (wxWindow::*m)(), const char *name);
};


template <typename W, typename... Args>
auto wxCreate(Args &&... args) -> decltype(wxCreator::create<W>(std::forward<Args>(args)...))
{
	return wxCreator::create<W>(std::forward<Args>(args)...);
}


class wxDialogQueue
{
	using creator_t = std::function<wxDialog *(wxWindow *parent)>;
	using opener_t = std::function<void (wxDialog *)>;

	struct entry_t
	{
		wxWeakRef<wxWindow> parent;
		creator_t creator;
		opener_t opener;
	};

	using stack = std::stack<wxWeakRef<wxDialog>>;
	using queue = std::queue<entry_t>;

	template <typename D, typename ...Args>
	struct Pusher
	{
		Pusher(wxDialogQueue &owner, wxWindow *parent, creator_t && creator)
			: owner_(owner)
			, parent_(parent)
			, creator_(std::move(creator))
		{
			wxASSERT_MSG(parent != nullptr, "parent must not be NULL");
		}

		~Pusher()
		{
			if (!opener_)
				opener_ = [](wxDialog *d) {	d->ShowModal();	};

			owner_.queue_.push({parent_, std::move(creator_), opener_});

			// Asynchronously invoke try_dequeue. The calling code must make no assumptions about whether pushing a dialog immediately opens it or not.
			owner_.timer_.CallAfter([owner = &this->owner_] {
				owner->try_dequeue();
			});
		}

		template <typename O>
		std::enable_if_t<std::is_invocable_v<O, D*>>
		operator|(O opener) &&
		{
			opener_ = [opener = std::move(opener)](wxDialog *d) {
				opener(static_cast<D*>(d));
			};
		}

		template <typename O>
		std::enable_if_t<std::is_invocable_v<O, D&>>
		operator|(O opener) &&
		{
			opener_ = [opener = std::move(opener)](wxDialog *d) {
				opener(*static_cast<D*>(d));
			};
		}

	private:
		wxDialogQueue &owner_;
		wxWindow *parent_;
		creator_t creator_;
		opener_t opener_;
	};

public:
	template <typename D, typename... Args>
		std::enable_if_t<std::is_base_of_v<wxDialog, D> && std::is_invocable_v<decltype(wxCreate<D, wxWindow *, Args...>), wxWindow *, Args...>,
	Pusher<D>> static push(wxWindow *parent, Args &&... args)
	{
		return {instance(), parent, [t = std::make_tuple(std::forward<Args>(args)...)](wxWindow *parent) {
			return std::apply([parent](auto &&... args) {
				return wxCreate<D>(parent, std::forward<decltype(args)>(args)...);
			}, std::move(t));
		}};
	}

private:
	wxDialogQueue();

	static wxDialogQueue &instance();
	void try_dequeue();
	bool must_be_delayed();

	stack stack_;
	queue queue_;

	wxTimer timer_;

#ifdef __WXMAC__
	std::stack<void *> shown_dialogs_creation_events_;
#endif
};

template <typename D = wxDialog, typename... Args, std::enable_if_t<!fx::trait::is_a_wxDialogEx_v<D>>* = nullptr>
auto wxPushDialog(Args &&... args) -> decltype(wxDialogQueue::push<wxDialogEx<D>>(std::forward<Args>(args)...))
{
	return wxDialogQueue::push<wxDialogEx<D>>(std::forward<Args>(args)...);
}

template <typename D = wxDialog, typename... Args, std::enable_if_t<fx::trait::is_a_wxDialogEx_v<D>>* = nullptr>
auto wxPushDialog(Args &&... args) -> decltype(wxDialogQueue::push<D>(std::forward<Args>(args)...))
{
	return wxDialogQueue::push<D>(std::forward<Args>(args)...);
}

template <typename W = wxPanel, typename... Args>
auto wxPage(wxBookCtrlBase *b, const wxString &title, bool select, Args &&... args) -> decltype(wxCreate<W>(b, std::forward<Args>(args)...))
{
	auto page = wxCreate<W>(b, std::forward<Args>(args)...);

	if (!dynamic_cast<wxSimplebook *>(b))
		b->SetInternalBorder(unsigned(wxDlg2Px(b, wxDefaultGap)));

	b->AddPage(page, title, select);
	return page;
}

template <typename W = wxPanel>
auto wxPage(wxBookCtrlBase *b, const wxString &title = wxEmptyString) -> decltype(wxCreate<W>(b))
{
	auto page = wxCreate<W>(b);

	if (!dynamic_cast<wxSimplebook *>(b))
		b->SetInternalBorder(unsigned(wxDlg2Px(b, wxDefaultGap)));

	b->AddPage(page, title);
	return page;
}

template <typename W>
struct wxTreebookPageWrapper
{
	wxTreebookPageWrapper(W *w, wxTreebook *b, wxPanel *p)
		: ctrl_(w)
		, book_(b)
		, page_(p)
	{}

	operator W*() const
	{
		return ctrl_;
	}

	W* operator->() const
	{
		return ctrl_;
	}

	W &operator*() const
	{
		return *ctrl_;
	}

	wxTreebook *GetBook() const
	{
		return book_;
	}

	int GetPagePos() const
	{
		return book_->FindPage(page_);
	}

	template <typename Func>
	std::enable_if_t<
		std::is_invocable_v<Func, wxTreebookPageWrapper &>,
		wxTreebookPageWrapper &&
	>
	operator |(Func && func) &&
	{
		std::forward<Func>(func)(*this);

		return std::move(*this);
	}

private:
	W *ctrl_;
	wxTreebook *book_;
	wxPanel *page_;
};

struct wxAutoRefocuser final
{
	wxAutoRefocuser();
	~wxAutoRefocuser();

private:
	int previously_focused_id_{wxNOT_FOUND};
};

template <typename P>
wxString wxTreebookSubPageTitle(wxTreebookPageWrapper<P> &parent, wxString title)
{
	auto book = parent.GetBook();
	auto page_pos = parent.GetPagePos();

	do {
		title.Prepend(wxS(" / ")).Prepend(book->GetPageText(page_pos));

		page_pos = book->GetPageParent(page_pos);
	} while (page_pos != wxNOT_FOUND);

	return title;
}

wxString wxGetContainingPageTitle(wxWindow *w, bool full_path = true);

template <typename W = wxPanel, typename... Args>
auto wxPage(wxTreebook *b, const wxString &title, bool select, Args &&... args) -> decltype(wxTreebookPageWrapper<W>(wxCreate<W>(std::declval<wxPanel*>(), std::forward<Args>(args)...), b, std::declval<wxPanel*>()))
{
	W *ctrl;

	auto page = new wxPanel(b) | [&] (auto p){
		wxVBox(p, 0) = {
			{ 0, wxTitle(p, title) },
			{ 1, ctrl = wxCreate<W>(p, std::forward<Args>(args)...) }
		};
	};

	b->SetInternalBorder(unsigned(wxDlg2Px(b, wxDefaultGap)));
	b->AddPage(page, title, select);
	return { ctrl, b, page };
}

template <typename W = wxPanel>
auto wxPage(wxTreebook *b, const wxString &title = wxEmptyString) -> decltype(wxTreebookPageWrapper<W>(wxCreate<W>(std::declval<wxPanel*>()), b, std::declval<wxPanel*>()))
{
	W *ctrl;

	auto page = new wxPanel(b) | [&] (auto p){
		wxVBox(p, 0) = {
			{ 0, wxTitle(p, title) },
			{ 1, ctrl = wxCreate<W>(p) }
		};
	};

	b->SetInternalBorder(unsigned(wxDlg2Px(b, wxDefaultGap)));
	b->AddPage(page, title);
	return { ctrl, b, page};
}

template <typename W = wxPanel, typename P, typename... Args, std::enable_if_t<
	!std::is_base_of_v<wxBookCtrlBase, P>
>* = nullptr>
auto wxPage(wxTreebookPageWrapper<P> &b, const wxString &title, bool select, Args &&... args) -> decltype(wxTreebookPageWrapper<W>(wxCreate<W>(std::declval<wxPanel*>(), std::forward<Args>(args)...), b.GetBook(), std::declval<wxPanel*>()))
{
	W* ctrl;
	wxTreebook *book = b.GetBook();

	auto page = new wxPanel(book) | [&] (auto p){
		wxVBox(p, 0) = {
			{ 0, wxTitle(p, wxTreebookSubPageTitle(b, title)) },
			{ 1, ctrl = wxCreate<W>(p, std::forward<Args>(args)...) }
		};
	};

	auto parent_page_pos = b.GetPagePos();

	book->InsertSubPage(parent_page_pos, page, title, select);
	book->ExpandNode(parent_page_pos, true);

	return { ctrl, book, page };
}

template <typename W = wxPanel, typename P, std::enable_if_t<
	!std::is_base_of_v<wxBookCtrlBase, P>
>* = nullptr>
auto wxPage(wxTreebookPageWrapper<P> &b, const wxString &title = wxEmptyString) -> decltype(wxTreebookPageWrapper<W>(wxCreate<W>(std::declval<wxPanel*>()), b.GetBook(), std::declval<wxPanel*>()))
{
	W* ctrl;
	wxTreebook *book = b.GetBook();

	auto page = new wxPanel(book) | [&] (auto p){
		wxVBox(p, 0) = {
			{ 0, wxTitle(p, wxTreebookSubPageTitle(b, title)) },
			{ 1, ctrl = wxCreate<W>(p) }
		};
	};

	auto parent_page_pos = b.GetPagePos();

	book->InsertSubPage(parent_page_pos, page, title);
	book->ExpandNode(parent_page_pos, true);

	return { ctrl, book, page };
}


/*************************/

wxWizardPageSimple *GetFirstPage(wxWizard *wiz);
int GetNumberOfFollowingPages(wxWizardPage *page);
int GetIndexOfPage(wxWizardPage *page);

template <typename W>
struct wxWizardPageWrapper
{
	wxWizardPageWrapper(W *w, wxWizard *wiz)
		: ctrl_(w)
		, wiz_(wiz)
	{}

	operator W*()
	{
		return ctrl_;
	}

	W* operator->()
	{
		return ctrl_;
	}

	W &operator*()
	{
		return *ctrl_;
	}

	wxWizard *GetWizard()
	{
		return wiz_;
	}


	template <typename Func>
	std::enable_if_t<
		std::is_invocable_v<Func, W*>,
		wxWizardPageWrapper &&
	>
	operator |(Func && func) &&
	{
		std::forward<Func>(func)(ctrl_);

		return std::move(*this);
	}

private:
	W *ctrl_;
	wxWizard *wiz_;
};

template <typename W = wxPanel, typename... Args>
auto wxPage(wxWizard *wiz, const wxString &title = wxEmptyString, bool = false, Args &&... args) -> W*//decltype(wxWizardPageWrapper<W>(wxCreate<W>(std::declval<wxPanel*>(), std::forward<Args>(args)...), wiz))
{
	W *ctrl;

	auto new_page = new wxWizardPageSimple(wiz) | [&] (auto p) {
		wxVBox(p, 0) = {
			{ 0, wxTitle(p, title) },
			{ wxSizerFlags(1).Expand(), ctrl = wxCreate<W>(p, std::forward<Args>(args)...) }
		};
	};

	if (auto last_page = GetFirstPage(wiz)) {
		while (last_page->GetNext())
			last_page = static_cast<wxWizardPageSimple *>(last_page->GetNext());

		last_page->Chain(new_page);
	}

	wiz->GetPageAreaSizer()->Add(new_page);

	//return { ctrl, wiz };
	return ctrl;
}


/*************************/


enum class wxValidateOnlyIfCurrent_t {};
inline constexpr wxValidateOnlyIfCurrent_t wxValidateOnlyIfCurrent{};

template <typename W = wxPanel>
auto wxPage(wxValidateOnlyIfCurrent_t) {
	return [](auto &&... args) -> decltype(wxPage<wxValidateOnlyIfCurrentPage<W>>(std::forward<decltype(args)>(args)...)) {
		return wxPage<wxValidateOnlyIfCurrentPage<W>>(std::forward<decltype(args)>(args)...);
	};
}


/*************************/

wxWindow *wxPageLink(wxWindow *parent, const wxString &label, wxBookCtrlBase *book, std::size_t pageid, int style = wxHL_DEFAULT_STYLE & ~(wxHL_CONTEXTMENU) );

/*************************/


void wxEnableStrictValidation(wxBookCtrlBase *b);

template <typename W, typename Func, std::enable_if_t<
	std::is_base_of_v<wxWindow, W> &&
	std::is_invocable_v<Func, W*>
>* = nullptr>
W* operator |(W *w, Func && func)
{
	if (w) {
		std::forward<Func>(func)(w);

		if constexpr (std::is_base_of_v<wxBookCtrlBase, W>)
			wxEnableStrictValidation(w);
	}

	return w;
}

template <typename W, typename Func, std::enable_if_t<std::conjunction_v<
	std::negation<std::is_pointer<W>>,
	std::is_base_of<wxWindow, std::decay_t<W>>,
	std::is_invocable<Func, std::decay_t<W>*>
>>* = nullptr>
decltype(auto) operator |(W && w, Func && func)
{
	std::forward<Func>(func)(&w);

	if constexpr (std::is_base_of_v<wxBookCtrlBase, std::decay_t<W>>)
		wxEnableStrictValidation(&w);

	return std::forward<W>(w);
}

/******************************/

wxSizer *wxSaveTextToFile(wxWindow *parent, wxTextCtrl &text, const wxString &label, const wxString &message = wxEmptyString, const wxString &default_name = wxEmptyString, const wxString &wildcards = wxT("*.txt"));
wxSizer *wxLoadTextFromFile(wxWindow *parent, wxTextCtrl &text, const wxString &label, const wxString &message = wxEmptyString, const wxString &default_name = wxEmptyString, const wxString &wildcards = wxT("*.txt"));
wxSizer *wxCopyTextToClipboard(wxWindow *parent, wxTextCtrl &text, const wxString &label);

wxSizer *wxSaveTextToFile(wxWindow *parent, std::function<wxString()> text_func, const wxString &label, const wxString &message = wxEmptyString, const wxString &default_name = wxEmptyString, const wxString &wildcards = wxT("*.txt"));
wxSizer *wxLoadTextFromFile(wxWindow *parent, std::function<void(const wxString &)> text_func, const wxString &label, const wxString &message = wxEmptyString, const wxString &default_name = wxEmptyString, const wxString &wildcards = wxT("*.txt"));

/*****************************/

class wxCheckBoxGroup: public wxPanel
{
public:
	struct CB
	{
		operator bool() const;
		CB &operator=(bool v);
		void Enable(bool enabled);
		void EnableAndSet(bool value);

	private:
		friend wxCheckBoxGroup;
		CB(wxCheckBox *cb);

		wxCheckBox *cb_;
	};

	wxCheckBoxGroup(wxWindow *parent);

	CB c(const wxString &label);

	bool IsAnyChecked() const;
	void SetValue(bool value);
	std::size_t GetCheckedNumber() const;
	std::size_t GetNumberOfEnabledCheckboxes() const;

private:
	wxHyperlinkCtrl *select_all_{};
	wxHyperlinkCtrl *deselect_all_{};
};

/*****************************/

// BUG: Clang requires these: clang's bug or something I don't understand about C++?
extern template struct wxValidateOnlyIfCurrentPage<wxChoicebook>;
extern template struct wxCreator::BookCtrlFix<wxValidateOnlyIfCurrentPage<wxChoicebook>>;

#endif // HELPERS_HPP
