#include <libfilezilla/glue/wx.hpp>

#include <wx/splitter.h>
#include <wx/msgdlg.h>
#include <wx/thread.h>
#include <wx/sizer.h>
#include <wx/log.h>
#include <wx/statline.h>
#include <wx/app.h>
#include <wx/hyperlink.h>
#include <wx/statbmp.h>
#include <wx/choicebk.h>
#include <wx/filepicker.h>
#include <wx/evtloop.h>
#include <wx/checkbox.h>

#include "../serveradministrator.hpp"
#include "../settings.hpp"
#include "../certinfoeditor.hpp"
#include "../settingsdialog.hpp"
#include "../networkconfigwizard.hpp"
#include "../textvalidatorex.hpp"

#include "../locale.hpp"
#include "../glue.hpp"
#include "../helpers.hpp"

void ServerAdministrator::ConfigureServer()
{
	if (!IsConnected())
		return;

	if (responses_to_await_for_ > 0) {
		logger_.log_u(fz::logmsg::debug_warning, L"Still retrieving server's configuration");
		return;
	}

	on_settings_received_func_ = [this] {
		logger_.log_u(fz::logmsg::status, _S("Server's configuration retrieved."));

		wxPushDialog<SettingsDialog>(this, fz::to_wxString(server_info_.name)) | [this](SettingsDialog &d) {
			on_settings_received_func_ = nullptr;

			settings_dialog_ = &d;

			settings_dialog_->SetServerPathFormat(server_path_format_);
			settings_dialog_->SetHostaddressAnyIsEquivalent(any_is_equivalent_);
			settings_dialog_->SetGroupsAndUsers(groups_, users_, server_can_impersonate_, server_username_);
			settings_dialog_->SetFilters(disallowed_ips_, allowed_ips_);
			settings_dialog_->SetProtocolsOptions(protocols_options_);
			settings_dialog_->SetFtpOptions(ftp_options_, ftp_tls_extra_info_);
			settings_dialog_->SetAdminOptions(admin_options_, admin_tls_extra_info_);
			settings_dialog_->SetLoggingOptions(logger_options_);
			settings_dialog_->SetAcmeOptions(acme_options_, acme_extra_account_info_);

			settings_dialog_->SetApplyFunction([this] {
				if (!client_.is_connected()) {
					wxMsgError(_S("Got disconnected from the server, cannot apply the changes.\n\nPlease, try again later."));
					return false;
				}

				auto && [new_groups, new_users] = settings_dialog_->GetGroupsAndUsers();
				auto && [new_disallowed_ips, new_allowed_ips] = settings_dialog_->GetFilters();
				auto && new_protocols_options = settings_dialog_->GetProtocolsOptions();
				auto && new_ftp_options = settings_dialog_->GetFtpOptions();
				auto && new_admin_options = settings_dialog_->GetAdminOptions();
				auto && new_logger_options = settings_dialog_->GetLoggingOptions();
				auto && new_acme_options = settings_dialog_->GetAcmeOptions();

				client_.send<administration::set_groups_and_users>(new_groups, new_users, true);
				client_.send<administration::set_ip_filters>(new_disallowed_ips, new_allowed_ips);
				client_.send<administration::set_protocols_options>(new_protocols_options);
				client_.send<administration::set_ftp_options>(new_ftp_options);
				client_.send<administration::set_admin_options>(new_admin_options);
				client_.send<administration::set_logger_options>(new_logger_options);
				client_.send<administration::set_acme_options>(new_acme_options);

				client_.send<administration::get_extra_certs_info>(CertInfoEditor::ftp_cert_id, new_ftp_options.sessions().tls.cert);
				client_.send<administration::get_extra_certs_info>(CertInfoEditor::admin_cert_id, new_admin_options.tls.cert);

				return true;
			});

			settings_dialog_->SetGenerateSelfsignedCertificateFunction([this](CertInfoEditor::cert_id id, const std::string &dn, const std::vector<std::string> &hostnames) {
				client_.send<administration::generate_selfsigned_certificate>(id, dn, hostnames);
			});

			settings_dialog_->SetGenerateAcmeCertificateFunction([this](CertInfoEditor::cert_id id, const fz::securable_socket::acme_cert_info &ci) {
				auto opts = settings_dialog_->GetAcmeOptions();
				logger_.log_u(fz::logmsg::status, _S("Generating ACME certificate"));
				client_.send<administration::generate_acme_certificate>(id, opts.how_to_serve_challenges, ci);
			});

			settings_dialog_->SetUploadCertificateFunction([this](CertInfoEditor::cert_id id, const std::string &cert, const std::string &key, const fz::native_string &password) {
				client_.send<administration::upload_certificate>(id, cert, key, password);
			});

			settings_dialog_->SetGenerateAcmeAccountFunction([this] { generate_acme_account(); });

			settings_dialog_->ShowModal();
			settings_dialog_ = nullptr;
		};
	};

	responses_to_await_for_ = 7;

	logger_.log_u(fz::logmsg::status, _S("Retrieving configuration from the server..."));

	client_.send<administration::get_groups_and_users>();
	client_.send<administration::get_ip_filters>();
	client_.send<administration::get_protocols_options>();
	client_.send<administration::get_ftp_options>();
	client_.send<administration::get_admin_options>();
	client_.send<administration::get_logger_options>();
	client_.send<administration::get_acme_options>();
}

void ServerAdministrator::ConfigureNetwork()
{
	if (!IsConnected())
		return;

	if (responses_to_await_for_ > 0) {
		logger_.log_u(fz::logmsg::debug_warning, L"Still retrieving server's configuration");
		return;
	}

	on_settings_received_func_ = [this] {
		wxPushDialog<NetworkConfigWizard>(this, _F("Network configuration wizard for server %s", fz::to_wxString(server_info_.name))) | [this](NetworkConfigWizard &wiz) {
			on_settings_received_func_ = nullptr;

			wiz.SetFtpOptions(ftp_options_);

			if (wiz.Run()) {
				if (!client_.is_connected()) {
					wxMsgError(_S("Got disconnected from the server, cannot apply the changes.\n\nPlease, try again later."));
					return;
				}

				client_.send<administration::set_ftp_options>(wiz.GetFtpOptions());
			}
		};
	};

	responses_to_await_for_ = 1;
	client_.send<administration::get_ftp_options>();
}
