#include <libfilezilla/glue/wx.hpp>

#include <cinttypes>
#include <array>
#include <numeric>
#include <vector>
#include <algorithm>

#include <wx/sizer.h>
#include <wx/menu.h>
#include <wx/renderer.h>
#include <wx/statline.h>

#include "sessionlist.hpp"
#include "locale.hpp"
#include "glue.hpp"
#include "helpers.hpp"

SessionList::SessionList()
{}

enum ListCol {
	Date,
	SessionID,
	Protocol,
	Username,
	Host,
	Status,
	Path,
	Size,
	InstSpeed,
	AvgSpeed,
	Progress
};

bool SessionList::Create(wxWindow *parent, wxWindowID winid, const wxPoint &pos, const wxSize &size,const wxString &name)
{
	if (!wxListCtrlEx::Create(parent, winid, pos, size, wxTAB_TRAVERSAL | wxLC_VIRTUAL | wxLC_REPORT, wxDefaultValidator, name))
		return false;

	auto cl = new wxFluidColumnLayoutManager(this);

	auto add_column = [&](int idx, auto name, size_t weight) {
		auto col = InsertColumn(idx, name);
		cl->SetColumnWeight(col, weight);
	};

	add_column(ListCol::Date,      _S("Date"),            1);
	add_column(ListCol::SessionID, _S("Session ID"),      1);
	add_column(ListCol::Protocol,  _S("Protocol"),        1);
	add_column(ListCol::Username,  _S("Username"),        1);
	add_column(ListCol::Host,      _S("Host"),            1);
	add_column(ListCol::Status,    _S("Transfer status"), 1);
	add_column(ListCol::Path,      _S("Path"),            3);
	add_column(ListCol::Size,      _S("Size"),            1);
	add_column(ListCol::InstSpeed, _S("Inst. Speed"),     0);
	add_column(ListCol::AvgSpeed,  _S("Avg. Speed"),      0);
	add_column(ListCol::Progress,  _S("Progress"),        2);

	Bind(wxEVT_CONTEXT_MENU, [&](const wxContextMenuEvent &ev){
		fz::scoped_lock lock(mutex_);

		long row = GetItemFromScreenPosition(ev.GetPosition());

		if (row < 0 || size_t(row) >= sessions_.size())
			return;

		auto session = &std::next(sessions_.cbegin(), row)->second;

		wxMenu menu;

		enum { Kill = 1, Ban, Info };

		menu.Append(Kill, _S("Kill session"));
		menu.Append(Ban, _S("Ban IP Address"));

		menu.Bind(wxEVT_MENU, [info = session->info, this](auto) {
			Event event(Event::SessionNeedsToBeKilled, info, *this);
			ProcessWindowEvent(event);
		}, Kill);

		menu.Bind(wxEVT_MENU, [info = session->info, this](auto) {
			Event event(Event::IpNeedsToBeBanned, info, *this);
			ProcessWindowEvent(event);
		}, Ban);

		if (session->info.sizer_creator) {
			menu.Append(Info, _S("More info..."));

			menu.Bind(wxEVT_MENU, [info = session->info, this](auto) mutable {
				show_session_info(std::move(info));
			}, Info);
		}

		PopupMenu(&menu);
	});

	Bind(wxEVT_LIST_CACHE_HINT, &SessionList::OnCacheHint, this);

	Bind(wxEVT_LIST_ITEM_ACTIVATED, [this](wxListEvent &ev) {
		ev.Skip();

		auto row = ev.GetItem().GetId();

		if (row < 0 || size_t(row) >= sessions_.size())
			return;

		auto info = std::next(sessions_.cbegin(), row)->second.info;
		show_session_info(std::move(info));
	});

	update_when_no_transfer_timer_.Bind(wxEVT_TIMER, [this](wxTimerEvent &) {
		DelayedUpdate();
	});

	return true;
}

void SessionList::OnCacheHint(wxListEvent &event)
{
	fz::scoped_lock lock(mutex_);

	update_when_no_transfer_timer_.Stop();

	cache_from_ = event.GetCacheFrom();
	cache_to_ = event.GetCacheTo();

	if (cache_from_ < 0 || size_t(cache_from_) >= sessions_.size()) {
		cache_from_ = -1;
		cache_to_ = -1;
		return;
	}

	auto amount = cache_to_ - cache_from_ +1;
	if (amount <= 0) {
		cache_from_ = -1;
		cache_to_ = -1;
		return;
	}

	cache_.clear();
	cache_.reserve(size_t(amount));

	bool transfers_ongoing = false;

	for (
		auto it = std::next(sessions_.cbegin(), cache_from_), end = sessions_.cend();
		amount > 0 && it != end;
		++it, --amount
	) {
		if (it->second.info.transfer.status != Idle)
			transfers_ongoing = true;

		cache_.push_back(it->second);
	}

	cache_to_ -= amount;

	cached_time_ = fz::monotonic_clock::now();
	if (transfers_ongoing)
		update_when_no_transfer_timer_.Start(1000);
}

SessionList::Session *SessionList::id2session(uint64_t id, bool emit_event)
{
	auto it = sessions_.find(id);
	if (it == sessions_.end()) {
		if (emit_event) {
			Session::Info info;
			info.id = id;

			GetEventHandler()->QueueEvent(new Event(Event::SessionNotAvailable, info, *this));
		}

		return nullptr;
	}

	return &it->second;
}

void SessionList::show_session_info(const Session::Info info)
{
	wxPushDialog(this->GetParent(), wxID_ANY, _F("Info about session id %d (%s)", info.id, info.host), wxDefaultPosition, wxDefaultSize, wxCAPTION | wxRESIZE_BORDER) | [info = std::move(info)](wxDialog *p) {
		wxVBox(p) = {
			wxGBox(p, 2, {1}) = {
				wxLabel(p, _S("Host: ")),
				wxEscapedLabel(p, info.host),

				wxLabel(p, _S("Session Id: ")),
				wxEscapedLabel(p, info.id_str),

				wxLabel(p, _S("Protocol: ")),
				wxEscapedLabel(p, info.protocol_name)
			},
			info.sizer_creator ? new wxStaticLine(p) : nullptr,
			info.sizer_creator ? info.sizer_creator(p, info) : nullptr,
			wxEmptySpace,
			new wxStaticLine(p),
			wxSizerFlags(0).Expand() >>= p->CreateStdDialogButtonSizer(wxOK),
		};

		p->ShowModal();
	};
}

wxString SessionList::OnGetItemText(long row, long column) const
{
	static const auto format_amount = [](long double amount) {
		wxString suffix;
		auto format = wxS("%.2Lf %s");

		if (amount < 1024) {
			suffix = _S("B");
			format = wxS("%.0Lf %s");
		}
		else
		if (amount < 1024*1024) {
			suffix = _S("KiB");
			amount /= 1024;
		}
		else
		if (amount < 1024*1024*1024) {
			suffix = _S("MiB");
			amount /= 1024*1024;
		}
		else {
			suffix = _S("GiB");
			amount /= 1024*1024*1024;
		}

		return wxString::Format(format, amount, suffix);
	};

	static const wxString second_suffix = wxS("/s");

	if (row >= 0 && row >= cache_from_ && row <= cache_to_) {
		auto *s = &cache_[size_t(row - cache_from_)];

		switch (column) {
			case ListCol::SessionID:
				return s->info.id_str;

			case ListCol::Protocol:
				return s->info.protocol_name;

			case ListCol::Username:
				return s->info.username;

			case ListCol::Host:
				return s->info.host;

			case ListCol::Date:
				return s->info.start_time_str;

			case ListCol::Status:
				switch (s->info.transfer.status) {
					case Uploading: return _S("uploading");
					case Downloading: return _S("downloading");
					case Idle: return _S("idle");
				}

				return wxEmptyString;

			case ListCol::Path:
				if (s->info.transfer.status != Idle)
					return s->info.transfer.path;

				return wxEmptyString;

			case ListCol::Size:
				if (s->info.transfer.status != Idle && s->info.transfer.size > 0)
					return format_amount((long double)s->info.transfer.size);

				return wxEmptyString;

			case ListCol::InstSpeed:
				if (s->info.transfer.status != Idle && s->info.transfer.inst_bytes_per_sec >= 0) {
					auto inst_bytes_per_sec = s->info.transfer.inst_bytes_per_sec;
					if ((cached_time_ - s->transfer_details.last_update).get_seconds() >= 2)
						inst_bytes_per_sec = 0;

					return format_amount((long double)inst_bytes_per_sec) + second_suffix;
				}

				return wxEmptyString;

			case ListCol::AvgSpeed:
				if (s->info.transfer.status != Idle && s->info.transfer.duration_millis >= 0) {
					auto duration_millis = s->info.transfer.duration_millis;
					auto delta_time_millis = (cached_time_ - s->transfer_details.last_update).get_milliseconds();

					if (delta_time_millis >= 2000)
						duration_millis += delta_time_millis;

					if (duration_millis > 0) {
						auto avg_bytes_per_sec = s->info.transfer.bytes_so_far*1000/duration_millis;

						return format_amount((long double)avg_bytes_per_sec) + second_suffix;
					}
				}

				return wxEmptyString;


			case ListCol::Progress:
				if (s->info.transfer.status == Downloading && s->info.transfer.progress >= 0)
					return wxString::Format(wxS("%d%%"), s->info.transfer.progress);

				return wxEmptyString;
		}
	}

	return wxEmptyString;
}

void SessionList::Add(uint64_t session_id, fz::datetime start_time, wxString host, int address_family)
{
	fz::scoped_lock lock(mutex_);

	auto [it, success] = sessions_.emplace(session_id, Session{});
	if (!success) return;
	auto &s = it->second;

	s.info.id = session_id;
	s.info.start_time = start_time;
	s.info.host = host;
	s.info.address_family = address_family;

	s.info.id_str = wxString::Format(wxS("%" PRIu64), s.info.id);
	s.info.start_time_str = fz::to_wxString(s.info.start_time);

	DelayedUpdate();
}

void SessionList::Remove(uint64_t session_id)
{
	fz::scoped_lock lock(mutex_);

	if (sessions_.erase(session_id) > 0)
		DelayedUpdate();
}

void SessionList::Clear()
{
	fz::scoped_lock lock(mutex_);

	if (!sessions_.empty())
		DelayedUpdate();

	sessions_.clear();
}

std::optional<SessionList::Session::Info> SessionList::GetInfo(uint64_t session_id)
{
	fz::scoped_lock lock(mutex_);

	if (auto *s = id2session(session_id, false))
		return s->info;

	return std::nullopt;
}

void SessionList::SetUserName(uint64_t session_id, wxString username)
{
	fz::scoped_lock lock(mutex_);

	auto *s = id2session(session_id, true);
	if (!s) return;

	s->info.username = std::move(username);

	DelayedUpdate();
}


void SessionList::SetTransferFileInfo(uint64_t session_id, fz::duration time, bool upload, wxString path, int64_t size, int64_t resume_offset, int64_t amount_at_start)
{
	fz::scoped_lock lock(mutex_);

	auto *sp = id2session(session_id, true);
	if (!sp) return;
	auto &s = *sp;

	s.info.transfer.start_time = time;
	s.info.transfer.status = upload ? Uploading : Downloading;
	s.info.transfer.path = std::move(path);
	s.info.transfer.size = size;
	s.info.transfer.rest = resume_offset > 0 ? resume_offset : 0;
	s.info.transfer.progress = size > 0 ? 0 : -1;

	s.transfer_details = { amount_at_start, 0, 0, fz::monotonic_clock::now() };

	SetTransferProgress(session_id, time, s.transfer_details.amount_at_start);
}

void SessionList::SetTransferProgress(uint64_t session_id, fz::duration time, int64_t amount)
{
	fz::scoped_lock lock(mutex_);

	auto *sp = id2session(session_id, true);
	if (!sp) return;
	auto &s = *sp;

	if (amount < 0) {
		s.info.transfer.status = Idle;
	}
	else {
		s.transfer_details.last_update = fz::monotonic_clock::now();

		s.info.transfer.duration_millis = (time - s.info.transfer.start_time).get_milliseconds();
		s.info.transfer.bytes_so_far = amount - s.transfer_details.amount_at_start;

		auto delta_duration_millis = s.info.transfer.duration_millis - s.transfer_details.previous_duration_millis;
		auto delta_bytes_so_far = s.info.transfer.bytes_so_far  - s.transfer_details.previous_bytes_so_far;
		s.info.transfer.inst_bytes_per_sec = delta_duration_millis > 0 ? delta_bytes_so_far*1000/delta_duration_millis : -1;

		s.transfer_details.previous_duration_millis = s.info.transfer.duration_millis;
		s.transfer_details.previous_bytes_so_far = s.info.transfer.bytes_so_far;

		auto total = s.info.transfer.rest + s.info.transfer.bytes_so_far;

		if (s.info.transfer.progress >= 0)
			s.info.transfer.progress = s.info.transfer.size > 0 ? (int)(total*100/s.info.transfer.size) : 0;

		if (s.info.transfer.status == Downloading) {
			s.info.total_uploaded = amount;

			if (s.info.transfer.progress == -1)
				s.info.transfer.size = total;
		}
		else
		if (s.info.transfer.status == Uploading) {
			s.info.total_downloaded = amount;
			s.info.transfer.size = total;
		}
	}

	DelayedUpdate();
}

void SessionList::SetProtocolInfo(uint64_t session_id, wxString name, SecureState state, Session::Info::SizerCreator info_sizer_creator)
{
	fz::scoped_lock lock(mutex_);

	auto *sp = id2session(session_id, true);
	if (!sp) return;
	auto &s = *sp;

	s.info.protocol_name = name;
	s.info.secure_state = state;
	s.info.sizer_creator = std::move(info_sizer_creator);
}

long SessionList::GetUpdatedItemCount() const
{
	fz::scoped_lock lock(mutex_);

	return static_cast<long>(sessions_.size());
}

wxDEFINE_EVENT(SessionList::Event::SessionNeedsToBeKilled, SessionList::Event);
wxDEFINE_EVENT(SessionList::Event::IpNeedsToBeBanned, SessionList::Event);
wxDEFINE_EVENT(SessionList::Event::SessionNotAvailable, SessionList::Event);

SessionList::Event::Event(wxEventType eventType, const Session::Info &session_info, const SessionList &owner)
	: wxEvent(owner.GetId(), eventType)
	, session_info(session_info)
{}
