#ifndef WXFZFTPSESSIONLIST_HPP
#define WXFZFTPSESSIONLIST_HPP

#include <libfilezilla/time.hpp>
#include <libfilezilla/mutex.hpp>

#include <map>
#include <memory>
#include <optional>

#include <wx/panel.h>
#include <wx/timer.h>

#include "listctrlex.hpp"
#include "fluidcolumnlayoutmanager.hpp"

class SessionList: public wxListCtrlEx
{
public:
	enum TransferStatus { Idle, Uploading, Downloading};
	enum SecureState { Insecure, QuasiSecure, Secure };

	struct Session {
		struct Info {
			using SizerCreator = std::function<wxSizer *(wxWindow *, const Info &)>;

			uint64_t id{};
			wxString id_str{};
			wxString username{};
			fz::datetime start_time{};
			wxString start_time_str{};
			wxString host{};
			int address_family{};
			wxString protocol_name{};
			SecureState secure_state{};
			SizerCreator sizer_creator;

			struct Transfer {
				TransferStatus status{};
				fz::duration start_time{};
				wxString path{};
				int64_t size{};
				int64_t rest{};

				int64_t duration_millis{};
				int64_t bytes_so_far{};
				int64_t inst_bytes_per_sec{};
				int progress{};
			} transfer;

			int64_t total_uploaded{};
			int64_t total_downloaded{};

			Info(){}
		} info{};

		struct TransferDetails {
			int64_t amount_at_start{};
			int64_t previous_bytes_so_far{};
			int64_t previous_duration_millis{};

			fz::monotonic_clock last_update{};
		} transfer_details{};

		Session(){}
	};

private:
	std::map<std::uint64_t, Session> sessions_{};

public:
	SessionList();

	bool Create(wxWindow *parent,
				wxWindowID winid = wxID_ANY,
				const wxPoint& pos = wxDefaultPosition,
				const wxSize& size = wxDefaultSize,
				const wxString& name = wxS("SessionList"));

	void Add(uint64_t session_id, fz::datetime start_time, wxString host, int address_family);
	void Remove(uint64_t session_id);
	void Clear();
	std::optional<SessionList::Session::Info> GetInfo(uint64_t session_id);

	void SetUserName(uint64_t session_id, wxString username);
	void SetTransferFileInfo(uint64_t session_id, fz::duration time, bool upload, wxString path, int64_t size, int64_t resume_offset, int64_t amount_at_start);
	void SetTransferProgress(uint64_t session_id, fz::duration time, int64_t amount);
	void SetProtocolInfo(uint64_t session_id, wxString name, SecureState state, Session::Info::SizerCreator info_sizer_creator);

	struct Event: wxEvent {
		virtual Event *Clone() const override {
			return new Event(*this);
		}

		const static wxEventTypeTag<Event> SessionNeedsToBeKilled;
		const static wxEventTypeTag<Event> IpNeedsToBeBanned;
		const static wxEventTypeTag<Event> SessionNotAvailable;

		Session::Info session_info;

	private:
		friend SessionList;

		Event(wxEventType eventType, const Session::Info &session_info, const SessionList &owner);
	};

private:
	long GetUpdatedItemCount() const override;
	wxString OnGetItemText(long item, long column) const override;
	void OnCacheHint(wxListEvent &event);

	mutable fz::mutex mutex_{true};

	Session *id2session(uint64_t id, bool emit_event);
	void show_session_info(const Session::Info info);

	std::vector<Session> cache_;
	long cache_from_{};
	long cache_to_{};
	fz::monotonic_clock cached_time_{};
	wxTimer update_when_no_transfer_timer_{};
};

#endif // WXFZFTPSESSIONLIST_HPP
