#include <libfilezilla/glue/wx.hpp>

#include <wx/sizer.h>
#include <wx/bookctrl.h>
#include <wx/treebook.h>
#include <wx/choicdlg.h>
#include <wx/textctrl.h>
#include <wx/checkbox.h>
#include <wx/display.h>
#include <wx/choicebk.h>
#include <wx/choice.h>
#include <wx/button.h>
#include <wx/log.h>
#include <wx/statline.h>
#include <wx/valgen.h>
#include <wx/simplebook.h>
#include <wx/radiobut.h>
#include <wx/scrolwin.h>

#include "passwordeditor.hpp"
#include "certinfoeditor.hpp"

#include "settingsdialog.hpp"
#include "groupseditor.hpp"
#include "userseditor.hpp"
#include "helpers.hpp"
#include "textvalidatorex.hpp"

#include "addressinfolisteditor.hpp"
#include "passwordeditor.hpp"
#include "certinfoeditor.hpp"
#include "filterseditor.hpp"
#include "integraleditor.hpp"

#include "locale.hpp"
#include "glue.hpp"

#include "../filezilla/util/bits.hpp"
#include "../filezilla/logger/type.hpp"

enum {
	impersonator_msw,
	impersonator_nix
};

static wxString address_and_port_in_other_editor(wxString address, unsigned int port, wxString other_address, const wxString &editor)
{
	if (address.empty())
		address = _F("Local port %d", port);
	else
		address = _F("Address %s", fz::join_host_and_port(address.ToStdWstring(), port));

	if (other_address.empty())
		other_address = _F("local port %d", port);
	else
		other_address = _F("address %s", fz::join_host_and_port(other_address.ToStdWstring(), port));

	if (address == other_address)
		return _F("%s is already used by %s", address, editor);

	return _F("%s conflicts with %s in %s", address, other_address, editor);
}

SettingsDialog::SettingsDialog()
{}

bool SettingsDialog::Create(wxWindow *parent, const wxString &server_name, wxWindowID winid, const wxPoint &pos, const wxSize &size, long style, const wxString &name)
{
	auto title = _F("Settings for server %s", server_name);

	if (!wxDialogEx::Create(parent, winid, title, pos, size, style, name))
		return false;

	server_name_ = fz::to_utf8(server_name);

	Bind(wxEVT_SHOW, [&](wxShowEvent &ev) {
		if (!ev.IsShown())
			return;

		check_expired_certs();
	});

	auto make_security_editor = [&](wxChoice *&min_tls_ver_ctrl, CertInfoEditor *&cert_info_ctrl, CertInfoEditor::cert_id cert_id) {
		return [&, cert_id](auto p) {
			return wxVBox(p) = {
				{ 0, wxHBox(p, 0) = {
					{ 0, wxLabel(p, _S("Minimum allowed TLS version:")) },
					{ 1, min_tls_ver_ctrl = new wxChoice(p, wxID_ANY)  | [&](wxChoice * p) {
						p->Append(_S("v1.2"));
						p->Append(_S("v1.3"));

						p->SetSelection(0);
					}},
				}},

				{ 0, new wxStaticLine(p, wxID_ANY) },

				{ 1, cert_info_ctrl = wxCreate<CertInfoEditor>(p, cert_id) }
			};
		};
	};

	static_cast<wxTreebook *>(GetBookCtrl()) | [&](auto p) {
		wxPage(p, _S("Server listeners")) | [&](auto p) {
			wxVBox(p, 0) = server_listeners_editor_ = wxCreate<AddressInfoListEditor>(p);
		};

		wxPage(p, _S("Protocols settings")) | [&](auto p) {
			wxVBox(p, 0) = wxCreate<wxNotebook>(p, wxID_ANY) | [&](auto p) {
				wxPage(p, _S("Autoban")) | [&](auto p) {
					wxVBox(p) = {
						{ 0, wxLabel(p, _S("&Ban IP address after:")) },
						{ 0, autoban_max_login_failures_ctrl_ = wxCreate<wxIntegralEditor>(p, _S("failed login attempts (0 disables the functionality)."), 1) },

						{ 0, wxLabel(p, _S("If they &happen within:")) },
						{ 0, autoban_login_failures_time_window_ctrl_ = wxCreate<wxIntegralEditor>(p, _S("seconds."), 1000) },

						{ 0, wxLabel(p, _S("The ban will &last:")) },
						{ 0, autoban_ban_duration_ctrl_ = wxCreate<wxIntegralEditor>(p, _S("seconds."), 1000) }
					};
				};

				wxPage(p,  _S("Timeouts")) | [&](auto p) {
					wxVBox(p) = {
						{ 0, wxLabel(p, _S("&Login timeout:")) },
						{ 0, login_timeout_ctrl_ = wxCreate<wxIntegralEditor>(p, _S("seconds."), 1000) },

						{ 0, wxLabel(p, _S("&Activity timeout:")) },
						{ 0, activity_timeout_ctrl_ = wxCreate<wxIntegralEditor>(p, _S("minutes."), 60000) },
					};
				};

				wxPage(p, _S("Performance")) | [&](auto p) {
					wxVBox(p) = {
						wxLabel(p, _S("Number of &threads:")),
						performance_number_of_session_threads_ctrl_ = wxCreate<wxIntegralEditor>(p),

						server_receiving_buffer_size_check_ = new wxCheckBox(p, wxID_ANY, _S("Customize TCP &receive buffer size for data socket:")),
						performance_receiving_buffer_size_ctrl_ = wxCreate<wxIntegralEditor>(p, _S("KiB"), 1024),

						server_sending_buffer_size_check_ = new wxCheckBox(p, wxID_ANY, _S("Customize TCP &send buffer size for data socket:")),
						performance_sending_buffer_size_ctrl_ = wxCreate<wxIntegralEditor>(p, _S("KiB"), 1024)
					};

					server_receiving_buffer_size_check_->Bind(wxEVT_CHECKBOX, [&](wxCommandEvent &ev) {
						performance_receiving_buffer_size_ctrl_->SetValue(ev.GetInt() ? 0 : -1);
					});

					server_sending_buffer_size_check_->Bind(wxEVT_CHECKBOX, [&](wxCommandEvent &ev) {
						performance_sending_buffer_size_ctrl_->SetValue(ev.GetInt() ? 0 : -1);
					});

					performance_receiving_buffer_size_ctrl_->Bind(wxIntegralEditor::Event::Changed, [&](wxIntegralEditor::Event &) {
						decltype(protocols_options_.performance.receive_buffer_size) size;

						bool enabled = performance_receiving_buffer_size_ctrl_->Get(size) && size != -1;

						performance_receiving_buffer_size_ctrl_->Enable(enabled);
						server_receiving_buffer_size_check_->SetValue(enabled);
					});

					performance_sending_buffer_size_ctrl_->Bind(wxIntegralEditor::Event::Changed, [&](wxIntegralEditor::Event &) {
						decltype(protocols_options_.performance.send_buffer_size) size;

						bool enabled = performance_sending_buffer_size_ctrl_->Get(size) && size != -1;

						performance_sending_buffer_size_ctrl_->Enable(enabled);
						server_sending_buffer_size_check_->SetValue(enabled);
					});
				};

				wxPage(p, _S("Filters")) |[&](auto p) {
					wxVBox(p) = filters_ctrl_ = wxCreate<FiltersEditor>(p);
				};
			};

			wxPage(p, _S("FTP and FTP over TLS (FTPS)")) | [&](auto p) {
				wxVBox(p, 0) = wxCreate<wxNotebook>(p, wxID_ANY) | [&](auto p) {
					wxPage(p, _S("Connection Security")) | make_security_editor(ftp_tls_min_ver_ctrl_, ftp_cert_info_ctrl_, CertInfoEditor::ftp_cert_id);

					wxPage(p, _S("Welcome message")) | [&](auto p) {
						wxVBox(p) = {
							{ 0, wxLabel(p, _S("Custom welcome message:"))},
							{ 1, welcome_message_ctrl_ = new wxTextCtrl(p, wxID_ANY, wxEmptyString, wxDefaultPosition, wxDefaultSize, wxTE_MULTILINE) },
						};

						auto font = welcome_message_ctrl_->GetFont();
						font.SetFamily(wxFONTFAMILY_TELETYPE);
						welcome_message_ctrl_->SetFont(font);
					};

					wxPage(p, _S("Passive mode")) | [&](auto p) {
						wxVBox(p) = {
							{ 0, use_custom_port_range_ctrl_ = new wxCheckBox(p, wxID_ANY, _S("&Use custom port range:")) },
							{ wxSizerFlags(0).Border(wxUP, 0), wxVBox(p) = {
								{ 0, wxLabel(p, _F("&From: (suggested is %d)", fz::port_randomizer::min_ephemeral_value)) },
								{ 0, min_port_range_ctrl_ = wxCreate<wxIntegralEditor>(p) },

								{ 0, wxLabel(p, _F("&To: (suggested is %d)", fz::port_randomizer::max_ephemeral_value)) },
								{ 0, max_port_range_ctrl_ = wxCreate<wxIntegralEditor>(p) }
							}},

							{ 0, wxLabel(p, _S("Use the following &host (leave empty to keep the default one):"))},
							{ 0, host_override_ctrl_ = new wxTextCtrl(p, wxID_ANY) },

							{ 0, disallow_host_override_for_local_peers_ctrl_ = new wxCheckBox(p, wxID_ANY, _S("Use the default host for local connections")) },
						};

						use_custom_port_range_ctrl_->Bind(wxEVT_CHECKBOX, [&](wxCommandEvent &ev) {
							bool enabled = ev.GetInt();

							auto &range = ftp_options_.sessions().pasv.port_range;

							if (enabled)
								range.emplace();
							else
								range.reset();

							min_port_range_ctrl_->Enable(enabled);
							min_port_range_ctrl_->SetRef(range->min, 1, 65535);

							max_port_range_ctrl_->Enable(enabled);
							max_port_range_ctrl_->SetRef(range->max, 1, 65535);
						});
					};
				};
			};
		};

		authentication_page_ = wxPage(p, _S("Rights management")) | [&](auto p) {
			wxRadioButton *radio_default{}, *radio_other{};
			wxPanel *panel_default{}, *panel_other{};
			wxWindow *label_default{};

			wxStaticVBox(p, _S("Default system user for filesystem access"), 0) = wxVBox(p) = {
				wxWText(p, _S("This setting controls which system user is used by the server for accessing files and directories. It applies to all server users that are not configured to use their own system credentials for accessing files and directories.")),
				wxWText(p, _S("The filesystem permissions of the system user selected here are checked for all file and directory operations. Newly created files and directories will have this system user listed as owner.")),

				radio_default = new wxRadioButton(p, 0, _S("Use system user the server is running under")),
				panel_default = new wxPanel(p) | [&] (auto p) {
					wxVBox(p) =	label_default = wxWText(p, _F("The server you are connected to is running as %s.", server_username_));
				},

				radio_other = new wxRadioButton(p, 1, _S("Use other system user")),
				panel_other = new wxPanel(p) | [&] (auto p) {
					wxVBox(p) =	default_impersonator_book_ = wxCreate<wxNavigationEnabled<wxSimplebook>>(p) | [&](auto p) {
						wxPage(wxValidateOnlyIfCurrent)(p, wxS("*MSW*")) |[&](auto p) {
							wxVBox(p, 0) = {
								{ 0, wxLabel(p, _S("Name:")) },
								{ 0, default_impersonator_name_msw_ = new wxTextCtrl(p, wxID_ANY) },

								{ 0, wxLabel(p, _S("Password:")) },
								{ 0, default_impersonator_password_msw_ = new wxTextCtrl(p, wxID_ANY, wxEmptyString, wxDefaultPosition, wxDefaultSize, wxTE_PASSWORD) },
							};
						};

						wxPage(wxValidateOnlyIfCurrent)(p, wxS("*NIX*")) |[&](auto p) {
							wxVBox(p, 0) = {
								{ 0, wxLabel(p, _S("Name:")) },
								{ 0, default_impersonator_name_nix_ = new wxTextCtrl(p, wxID_ANY) },

								{ 0, wxLabel(p, _S("Group (optional, leave blank to pick the default one):")) },
								{ 0, default_impersonator_group_nix_ = new wxTextCtrl(p, wxID_ANY) },
							};
						};
					};
				},

				wxWText(p, _S("Access rights for all server users are further restricted by their configured mounts and permissions."))
			};

			auto set_selected = [v = std::vector<std::pair<wxRadioButton *, wxPanel *>>{{radio_default, panel_default}, {radio_other, panel_other}}](std::size_t n) {
				if (n >= v.size())
					return;

				for (auto i = std::size_t{}, end_i = v.size(); i < end_i;  ++i) {
					auto enabled = i == n;
					v[i].first->SetValue(enabled);
					v[i].second->Enable(enabled);
				}
			};

			p->Bind(wxEVT_RADIOBUTTON, [set_selected](wxCommandEvent &evt) {
				set_selected(std::size_t(evt.GetId()));
			});

			wxTransferDataToWindow(p, [this, p, label_default, radio_other, set_selected = std::move(set_selected)] {
				bool can_impersonate = can_impersonate_;

				if (auto imp = users_.default_impersonator.msw()) {
					default_impersonator_book_->ChangeSelection(impersonator_msw);

					radio_other->SetValidator(wxGenericValidator(&imp->enabled));
					set_selected(imp->enabled);

					default_impersonator_name_msw_->SetValidator(TextValidatorEx(wxFILTER_NONE, &imp->name, field_must_not_be_empty(_S("Name"))));
					default_impersonator_password_msw_->SetValidator(TextValidatorEx(wxFILTER_NONE, &imp->password, field_must_not_be_empty(_S("Password"))));
				}
				else
				if (auto imp = users_.default_impersonator.nix()) {
					default_impersonator_book_->ChangeSelection(impersonator_nix);

					radio_other->SetValidator(wxGenericValidator(&imp->enabled));
					set_selected(imp->enabled);

					default_impersonator_name_nix_->SetValidator(TextValidatorEx(wxFILTER_NONE, &imp->name, field_must_not_be_empty(_S("Name"))));
					default_impersonator_group_nix_->SetValidator(TextValidatorEx(wxFILTER_NONE, &imp->group));
				}
				else
					can_impersonate = false;

				label_default->SetLabel(_F("The server you are connected to is running as %s.", server_username_));
				label_default->Refresh();

				p->Enable(can_impersonate);
				return true;
			});

			groups_editor_ = wxPage<GroupsEditor>(p, _S("Groups"));

			users_editor_ = wxPage<UsersEditor>(p, _S("Users"));
		};

		wxPage(p, _S("Administration"), false, wxID_ANY) | [&](auto p) {
			wxVBox(p, 0) = wxCreate<wxNotebook>(p, wxID_ANY) | [&](auto p) {
				wxPage(p,  _S("Connection")) | [&](auto p) {
					#if defined ENABLE_ADMIN_DEBUG && ENABLE_ADMIN_DEBUG
						bool const edit_tls_mode = true;
					#else
						bool const edit_tls_mode = false;
					#endif

					wxVBox(p) = {
						{0, wxLabel(p, _S("&Local listening port:")) },
						{0, admin_port_ctrl_ = wxCreate<wxIntegralEditor>(p) },

						{0, wxLabel(p, _S("&Password:")) },
						{0, admin_password_ctrl_ = wxCreate<PasswordEditor>(p) },

						{0, wxLabel(p, _S("&Additional listeners (only available if a password is set):")) },
						{1, admin_listeners_editor_ = wxCreate<AddressInfoListEditor>(p, edit_tls_mode) }
					};
				};

				wxPage(p,  _S("Connection security")) | make_security_editor(admin_tls_min_ver_ctrl_, admin_cert_info_ctrl_, CertInfoEditor::admin_cert_id);
			};

			admin_password_ctrl_->Bind(PasswordEditor::Event::Changed, [&](PasswordEditor::Event &) {
				admin_listeners_editor_->Enable(admin_password_ctrl_->HasPassword());
			});

			admin_listeners_editor_->Disable();
		};

		wxPage(p, _S("Logging")) | [&](auto p) {
			wxVBox(p, 0) = {
				{ 0, wxLabel(p,  _S("Logging &level (the higher the more verbose):")) },
				{ 0, log_level_ctrl_ = new wxChoice(p, wxID_ANY) | [&](wxChoice * p) {
					p->Append(_S("0 - None"));
					p->Append(_S("1 - Normal"));
					p->Append(_S("2 - Warning"));
					p->Append(_S("3 - Info"));
					p->Append(_S("4 - Verbose"));
					p->Append(_S("5 - Debug"));

					p->SetSelection(0);
				}},

				{ 0, wxLabel(p,  _S("Log &output:")) },
				{ 0, logging_choice_ctrl_ = new wxChoicebook(p, wxID_ANY) |[&](wxChoicebook *p) {
					wxPage(wxValidateOnlyIfCurrent)(p, _S("To file")) |[&](auto p) {
						wxVBox(p, 0) = {
							wxLabel(p, _S("File &path (must be absolute):")),
							log_path_ctrl_ = new wxTextCtrl(p, wxID_ANY),

							log_rotation_choice_ctrl_ = new wxChoicebook(p, wxID_ANY) |[&](wxChoicebook *p) {
								wxPage(wxValidateOnlyIfCurrent)(p, _S("File rotation is disabled")) | [&](auto p) {
									wxTransferDataFromWindow(p, [this] {
										logger_opts_.max_amount_of_rotated_files(0);
										return true;
									});
								};

								wxPage(wxValidateOnlyIfCurrent)(p, _S("File rotation is enabled")) | [&](auto p) {
									wxVBox(p, 0) = {
										wxLabel(p, _S("Maximum number of &log files:")),
										log_rotations_amount_ctrl_ = wxCreate<wxIntegralEditor>(p),

										wxLabel(p, _S("Log rotation &policy:")),
										log_rotation_type_ctrl_ = new wxChoicebook(p, wxID_ANY) |[&](wxChoicebook *p) {
											wxPage(wxValidateOnlyIfCurrent)(p, _S("Rotate log file when its size reaches the maximum size")) | [&](auto p) {
												wxVBox(p, 0) = {
													wxLabel(p, _S("Maximum file &size:")),
													log_file_size_ctrl_ = wxCreate<wxIntegralEditor>(p, _S("MiB"), 1024*1024)
												};
											};

											wxPage(wxValidateOnlyIfCurrent)(p, _S("Rotate log files daily"));

											wxTransferDataFromWindow(p, [p, this] {
												logger_opts_.rotation_type(fz::logger::file::options::rotation_type_t(p->GetSelection()));
												return true;
											});
										}
									};
								};
							},
						};

						wxTransferDataFromWindow(p, [&] {
							logger_opts_.include_headers(true);
							return true;
						});
					};

					wxPage(wxValidateOnlyIfCurrent)(p, _S("To standard error"), true) |[&](auto p) {
						log_include_headers_ctrl_ = new wxCheckBox(p, wxID_ANY, _S("Log lines are timestamped."));

						wxTransferDataFromWindow(p, [&] {
							logger_opts_.name().clear();
							return true;
						});
					};

					p->Disable();
				}}
			};
		};

		auto acme_page = wxPage(p, _S("Let's Encrypt®")) | [&](auto p) {
			wxPanel *acme_panel{};

			wxVBox(p, 0) = {
				{ 0, acme_enabled_ = new wxCheckBox(p, wxID_ANY, _S("Enable Let's Encrypt® certificate generation")) },
				{ 1, acme_panel = new wxPanel(p) | [&] (auto p) {
					wxVBox(p, 0) = {
						{ 0, wxStaticVBox(p, _S("Default account")) = wxGBox(p, 2, {1}) = {
							{ 0, wxLabel(p,  _S("ID:")) },
							{ 0, acme_account_id_ctrl_ = new wxTextCtrl(p, wxID_ANY,  wxEmptyString, wxDefaultPosition, wxDefaultSize, wxTE_READONLY) },

							{ 0, wxLabel(p,  _S("Directory:")) },
							{ 0, new wxPanel(p) | [&](auto p) {
								wxVBox(p, 0) = acme_account_directory_ctrl_ = new wxTextCtrl(p, wxID_ANY,  wxEmptyString, wxDefaultPosition, wxDefaultSize, wxTE_READONLY);
								wxTransferDataFromWindow(p, [this] {
									if (acme_account_directory_ctrl_->GetValue().IsEmpty() && acme_enabled_->GetValue()) {
										wxMsgError(_S("You must create an account, or disable Let's Encrypt® certificate generation."));
										return false;
									}

									return true;
								});
							}},
							{ 0, wxLabel(p,  _S("Contacts:")) },
							{ 0, acme_account_contacts_ctrl_ = new wxTextCtrl(p, wxID_ANY,  wxEmptyString, wxDefaultPosition, wxDefaultSize, wxTE_READONLY) }
						}},
						{ 0, acme_account_id_button_ = new wxButton(p, wxID_ANY, _S("Create new account")) },

						{ 0, wxLabel(p,  _S("How to &perform the challenges:")) },
						{ 1, acme_how_ctrl_ = new wxChoicebook(p, wxID_ANY) | [&](wxChoicebook * p) {
							wxPage(wxValidateOnlyIfCurrent)(p, _S("Please, make a choice")) |[&](auto p) {
								wxTransferDataFromWindow(p, [this] {
									if (acme_enabled_->GetValue()) {
										wxMsgError(_S("You must choose how to perform a challenge"));
										return false;
									}

									acme_opts_.how_to_serve_challenges = {};
									return true;
								});
							};

							wxPage(wxValidateOnlyIfCurrent)(p, _S("Use an internal web server")) |[&] (auto p) {
								wxVBox(p, 0) = {
									{0, wxLabel(p, _S("&Listeners:")) },
									{1, acme_listeners_ctrl_ = wxCreate<AddressInfoListEditor>(p, false) },
								};

								wxTransferDataFromWindow(p, [this] {
									if (acme_listeners_.empty()) {
										wxMsgError(_S("You must add at least one listener"));
										acme_listeners_ctrl_->SetFocus();
										return false;
									}

									acme_opts_.how_to_serve_challenges = fz::acme::serve_challenges::internally{acme_listeners_};

									return true;
								});

								wxTransferDataToWindow(p, [this] {
									if (!acme_listeners_.empty())
										return true;

									if (std::holds_alternative<fz::acme::serve_challenges::internally>(acme_opts_.how_to_serve_challenges))
										return true;

									acme_listeners_.push_back({"0.0.0.0", 80});
									acme_listeners_.push_back({"::", 80});
									acme_listeners_ctrl_->SetLists({
										{&acme_listeners_, {{fz::tcp::address_info::allow_tls, _S("HTTP")}}}
									});

									return true;
								});
							};

							wxPage(wxValidateOnlyIfCurrent)(p, _S("Use an external web server")) |[&] (auto p) {
								wxVBox(p, 0) = {
									{ 0, wxLabel(p,  _S("&Path from which the web server responds to GET requests to \"/.well-known/acme-challenge/\":")) },
									{ 0, acme_well_knon_path_ctrl_ = new wxTextCtrl(p, wxID_ANY) },

									{ 0, acme_create_path_if_not_exist_ctrl_ = new wxCheckBox(p, wxID_ANY, _S("&Create path if it doesn't exist already.")) },
								};

								wxTransferDataFromWindow(p, [this] {
									acme_opts_.how_to_serve_challenges = fz::acme::serve_challenges::externally {
										acme_well_known_path_,
										acme_create_path_if_not_exist_ctrl_->GetValue()
									};

									return true;
								});
							};

							wxTransferDataFromWindow(p, [this] {
								ftp_cert_info_ctrl_->SetAcmeOptions(acme_opts_);
								//admin_cert_info_ctrl_->SetAcmeOptions(acme_opts_);

								return true;
							});
						}},

						{ 0, new wxStaticLine(p, wxID_ANY) },
						{ 0, wxLabel(p, _S("Let's Encrypt® is a trademark of the Internet Security Research Group. All rights reserved.")) |[&](wxWindow *p) {
							p->SetFont(p->GetFont().MakeItalic().MakeSmaller());
						}}
					};
				}
			}};

			acme_account_id_button_->Bind(wxEVT_BUTTON, [this](auto) mutable {
				if (acme_generate_account_func_) {
					acme_account_id_button_->Disable();

					acme_account_id_ctrl_->SetValue(_S("Generating..."));
					acme_account_directory_ctrl_->SetValue(_S("Generating..."));
					acme_account_contacts_ctrl_->SetValue(_S("Generating..."));
					acme_account_id_button_->SetLabel(_S("Generating..."));

					acme_account_id_ctrl_->Disable();
					acme_account_directory_ctrl_->Disable();
					acme_account_contacts_ctrl_->Disable();
					acme_account_id_button_->Disable();

					acme_generate_account_func_();
				}
			});

			acme_enabled_->Bind(wxEVT_CHECKBOX, [this, acme_panel] (const wxCommandEvent &ev) {
				if (ev.GetInt()) {
					acme_panel->Enable();
				}
				else {
					acme_panel->Disable();
					acme_how_ctrl_->ChangeSelection(0);
				}
			});
		};

		groups_editor_->Bind(GroupsList::Event::Changed, [this] (GroupsList::Event &) {
			users_editor_->SetGroups(groups_);
		});

		groups_editor_->Bind(GroupsList::Event::Changing, &SettingsDialog::OnGroupsListChanging, this);

		ftp_cert_info_ctrl_->SetSwitchToAcmeOptsFunc([p, pos = acme_page.GetPagePos()] {
			p->ChangeSelection(size_t(pos));
		});

		admin_cert_info_ctrl_->SetSwitchToAcmeOptsFunc([p, pos = acme_page.GetPagePos()] {
			p->ChangeSelection(size_t(pos));
		});
	};

	CreateButtons(wxAPPLY | wxOK | wxCANCEL);

	LayoutDialog();

	auto parent_display_id = wxDisplay::GetFromWindow(parent);
	wxDisplay display(parent_display_id == wxNOT_FOUND ? 0 : unsigned(parent_display_id));
	auto area = display.GetClientArea();

	auto min_size = GetMinSize();

	if (auto h = area.GetHeight()/2; min_size.GetHeight() < h)
		min_size.SetHeight(h);
	if (auto w = area.GetWidth()/2; min_size.GetWidth() < w)
		min_size.SetWidth(w);

	SetMinSize(min_size);
	SetSize(min_size);

	return true;
}

void SettingsDialog::SetServerPathFormat(fz::util::fs::path_format server_path_format)
{
	server_path_format_ = server_path_format;
}

void SettingsDialog::SetHostaddressAnyIsEquivalent(bool any_is_equivalent)
{
	any_is_equivalent_ = any_is_equivalent;
}

void SettingsDialog::SetGroupsAndUsers(const fz::authentication::file_based_authenticator::groups &groups, const fz::authentication::file_based_authenticator::users &users, bool can_impersonate, const std::wstring server_username)
{
	groups_ = groups;
	users_ = users;
	can_impersonate_ = can_impersonate;
	server_username_ = server_username;

	groups_editor_->SetGroups(groups_);
	users_editor_->SetGroups(groups_);
	users_editor_->SetUsers(users_, server_name_);

	groups_editor_->SetServerPathFormat(server_path_format_);
	users_editor_->SetServerPathFormat(server_path_format_);

	authentication_page_->TransferDataToWindow();
}

std::pair<const fz::authentication::file_based_authenticator::groups &, const fz::authentication::file_based_authenticator::users &>
SettingsDialog::GetGroupsAndUsers() const
{
	return {groups_, users_};
}

void SettingsDialog::SetProtocolsOptions(const server_settings::protocols_options &protocols_options)
{
	protocols_options_ = protocols_options;

	autoban_max_login_failures_ctrl_->SetRef(protocols_options_.autobanner.max_login_failures());
	autoban_login_failures_time_window_ctrl_->SetRef(protocols_options_.autobanner.login_failures_time_window(), fz::duration::from_milliseconds(0));
	autoban_ban_duration_ctrl_->SetRef(protocols_options_.autobanner.ban_duration(), fz::duration::from_milliseconds(0));

	performance_number_of_session_threads_ctrl_->SetRef(protocols_options_.performance.number_of_session_threads, 0, 256);
	performance_receiving_buffer_size_ctrl_->SetRef(protocols_options_.performance.receive_buffer_size, -1)->set_mapping({{-1, _S("Use default")}});
	performance_sending_buffer_size_ctrl_->SetRef(protocols_options_.performance.send_buffer_size, -1)->set_mapping({{-1, _S("Use default")}});

	login_timeout_ctrl_->SetRef(protocols_options_.timeouts.login_timeout, fz::duration::from_milliseconds(0));
	activity_timeout_ctrl_->SetRef(protocols_options_.timeouts.activity_timeout, fz::duration::from_milliseconds(0));
}

const server_settings::protocols_options &SettingsDialog::GetProtocolsOptions() const
{
	return protocols_options_;
}

void SettingsDialog::SetFtpOptions(const fz::ftp::server::options &ftp_options, const fz::securable_socket::cert_info::extra &tls_extra_info)
{
	ftp_options_ = ftp_options;
	ftp_tls_extra_info_ = tls_extra_info;

	ftp_tls_min_ver_ctrl_->SetSelection(std::min(1, std::max(0, int(ftp_options_.sessions().tls.min_tls_ver) - int(fz::tls_ver::v1_2))));
	ftp_cert_info_ctrl_->SetValue(&ftp_options_.sessions().tls.cert, tls_extra_info);
	server_listeners_editor_->SetLists({
		{&ftp_options_.listeners_info(), {
			{ fz::tcp::address_info::allow_tls, _S("Explicit FTP over TLS and insecure plain FTP") },
			{ fz::tcp::address_info::implicit_tls, _S("Implicit FTP over TLS (deprecated)") },
			{ fz::tcp::address_info::require_tls, _S("Require explicit FTP over TLS") }
		}},
	});

	auto &range = ftp_options_.sessions().pasv.port_range;

	use_custom_port_range_ctrl_->SetValue(range.has_value());
	if (range.has_value()) {
		min_port_range_ctrl_->Enable();
		min_port_range_ctrl_->SetRef(range->min, 1, 65535);

		max_port_range_ctrl_->Enable();
		max_port_range_ctrl_->SetRef(range->max, 1, 65535);
	}
	else {
		min_port_range_ctrl_->Disable();
		min_port_range_ctrl_->SetRef(nullptr);

		max_port_range_ctrl_->Disable();
		max_port_range_ctrl_->SetRef(nullptr);
	}

	host_override_ctrl_->SetValidator(TextValidatorEx(wxFILTER_NONE, &ftp_options_.sessions().pasv.host_override));
	disallow_host_override_for_local_peers_ctrl_->SetValidator(wxGenericValidator(&ftp_options_.sessions().pasv.do_not_override_host_if_peer_is_local));

	welcome_message_ctrl_->SetValidator(TextValidatorEx(wxFILTER_NONE, &ftp_options_.welcome_message(), [this](const wxString &str, bool is_final_validation) {
		wxString msg;

		if (is_final_validation) {
			auto welcome = fz::ftp::commander::welcome_message_t(fz::to_utf8(str));
			if (welcome.empty() && !str.empty())
				return _S("Text contains invalid characters");

			auto res = welcome.validate();

			switch (res) {
				case decltype(res)::ok:
					break;

				case decltype(res)::line_too_long:
					msg = _F("Line cannot take up more than %zu bytes", decltype(res)::line_limit);
					break;

				case decltype(res)::total_size_too_big:
					msg = _F("Total size of message cannot exceeed %zu bytes", decltype(res)::total_limit);
					break;
			}

			if (!res) {
				auto utf8_from = res.data().data() - welcome.data();
				auto wstring_from = fz::to_wstring_from_utf8(welcome.substr(0, std::size_t(utf8_from))).size();
				auto wstring_to = wstring_from + fz::to_wstring_from_utf8(res.data()).size();

				welcome_message_ctrl_->CallAfter([wstring_from, wstring_to, this] {
					welcome_message_ctrl_->SetSelection(long(wstring_from), long(wstring_to));
					welcome_message_ctrl_->ShowPosition(long(wstring_from));
				});
			}
		}

		return msg;
	}));
}

const fz::ftp::server::options &SettingsDialog::GetFtpOptions() const
{
	return ftp_options_;
}

void SettingsDialog::SetAdminOptions(const server_settings::admin_options &admin_options, const fz::securable_socket::cert_info::extra &tls_extra_info)
{
	admin_options_ = admin_options;
	admin_tls_extra_info_ = tls_extra_info;

	admin_port_ctrl_->SetRef(admin_options_.local_port, 1, 65535);
	admin_password_ctrl_->SetPassword(&admin_options_.password);
	admin_listeners_editor_->SetLists({
		{ &admin_options_.additional_address_info_list, {
			{ fz::tcp::address_info::implicit_tls, _S("Administration") }
		}}
	});

	admin_tls_min_ver_ctrl_->SetSelection(std::min(1, std::max(0, int(admin_options_.tls.min_tls_ver) - int(fz::tls_ver::v1_2))));
	admin_cert_info_ctrl_->SetValue(&admin_options_.tls.cert, tls_extra_info);

	admin_listeners_editor_->Enable(admin_options_.password.has_password());
}

const server_settings::admin_options &SettingsDialog::GetAdminOptions() const
{
	return admin_options_;
}

void SettingsDialog::SetFilters(const fz::tcp::binary_address_list &disallowed_ips, const fz::tcp::binary_address_list &allowed_ips)
{
	disallowed_ips_ = disallowed_ips;
	allowed_ips_ = allowed_ips;

	filters_ctrl_->SetIps(&disallowed_ips_, &allowed_ips_);
}

std::pair<const fz::tcp::binary_address_list &, const fz::tcp::binary_address_list &> SettingsDialog::GetFilters() const
{
	return {disallowed_ips_, allowed_ips_};
}

void SettingsDialog::SetLoggingOptions(const fz::logger::file::options &opts)
{
	logger_opts_ = opts;

	logging_choice_ctrl_->Enable();
	logging_choice_ctrl_->SetSelection(opts.name().empty());
	log_rotation_choice_ctrl_->SetSelection(opts.max_amount_of_rotated_files() > 0);
	log_rotation_type_ctrl_->SetSelection(opts.rotation_type());

	SetLogTypes(logger_opts_.enabled_types());
	log_level_old_selection_ = log_level_ctrl_->GetSelection();

	log_path_ctrl_->SetValidator(TextValidatorEx(wxFILTER_NONE, &logger_opts_.name(), absolute_path_validation(server_path_format_)));
	log_rotations_amount_ctrl_->SetRef(logger_opts_.max_amount_of_rotated_files(), 1, 10000);
	log_file_size_ctrl_->SetRef(logger_opts_.max_file_size(), 1*1024*1024)->set_mapping({
		{ logger_opts_.max_possible_size, _S("Maximum possible") }
	});
	log_include_headers_ctrl_->SetValidator(wxGenericValidator(&logger_opts_.include_headers()));
}

const fz::logger::file::options &SettingsDialog::GetLoggingOptions() const
{
	return logger_opts_;
}

void SettingsDialog::SetAcmeAccountId(const std::string &account_id, const fz::acme::extra_account_info &extra)
{
	acme_opts_.account_id = account_id;
	acme_extra_account_info_ = extra;

	if (!extra)
		acme_opts_.account_id.clear();

	acme_account_id_button_->Enable();
	acme_account_id_button_->SetLabel(_S("Generate new account"));

	if (acme_opts_.account_id.empty()) {
		auto font = acme_account_id_ctrl_->GetFont();
		font.SetStyle(wxFONTSTYLE_ITALIC);

		acme_account_id_ctrl_->SetFont(font);
		acme_account_id_ctrl_->SetValue(_S("<Account not created yet>"));
	}
	else {
		auto font = acme_account_id_ctrl_->GetFont();
		font.SetStyle(wxFONTSTYLE_NORMAL);

		acme_account_id_ctrl_->SetFont(font);
		acme_account_id_ctrl_->SetValue(fz::to_wxString(account_id));
	}

	acme_account_directory_ctrl_->SetValue(fz::to_wxString(extra.directory));
	acme_account_contacts_ctrl_->SetValue(fz::join<wxString>(extra.contacts));

	acme_account_id_ctrl_->Enable();
	acme_account_directory_ctrl_->Enable();
	acme_account_contacts_ctrl_->Enable();
	acme_account_id_button_->Enable();
}

void SettingsDialog::SetAcmeOptions(const server_settings::acme_options &acme_options, const fz::acme::extra_account_info &extra)
{
	acme_opts_ = acme_options;
	acme_extra_account_info_ = extra;

	if (!extra)
		acme_opts_.account_id.clear();

	SetAcmeAccountId(acme_opts_.account_id, extra);

	ftp_cert_info_ctrl_->SetAcmeOptions(acme_opts_);
	//admin_cert_info_ctrl_->SetAcmeOptions(acme_opts_);

	if (auto h = std::get_if<fz::acme::serve_challenges::internally>(&acme_opts_.how_to_serve_challenges)) {
		acme_how_ctrl_->ChangeSelection(1);
		acme_listeners_ = h->addresses_info;
	}
	else
	if (auto h = std::get_if<fz::acme::serve_challenges::externally>(&acme_opts_.how_to_serve_challenges)) {
		acme_how_ctrl_->ChangeSelection(2);
		acme_well_known_path_ = h->well_known_path;
		acme_create_path_if_not_exist_ctrl_->SetValue(h->create_if_not_existing);
		acme_listeners_.clear();
	}
	else
		acme_how_ctrl_->ChangeSelection(0);

	acme_enabled_->SetValue(acme_opts_.how_to_serve_challenges && extra);
	auto evt = wxCommandEvent(wxEVT_CHECKBOX);
	evt.SetInt(acme_opts_.how_to_serve_challenges && extra);
	acme_enabled_->GetEventHandler()->ProcessEvent(evt);

	acme_well_knon_path_ctrl_->SetValidator(TextValidatorEx(wxFILTER_NONE, &acme_well_known_path_, absolute_path_validation(server_path_format_)));

	acme_listeners_ctrl_->SetLists({
		{&acme_listeners_, {{fz::tcp::address_info::allow_tls, _S("HTTP")}}}
	});
}

const server_settings::acme_options &SettingsDialog::GetAcmeOptions()
{
	return acme_opts_;
}

void SettingsDialog::SetApplyFunction(std::function<bool ()> func)
{
	apply_function_ = std::move(func);
}

void SettingsDialog::SetGenerateSelfsignedCertificateFunction(CertInfoEditor::GenerateSelfsignedFunc func)
{
	ftp_cert_info_ctrl_->SetGenerateSelfsignedCertificateFunction(func);
	admin_cert_info_ctrl_->SetGenerateSelfsignedCertificateFunction(std::move(func));
}

void SettingsDialog::SetGenerateAcmeCertificateFunction(CertInfoEditor::GenerateAcmeFunc func)
{
	ftp_cert_info_ctrl_->SetGenerateAcmeCertificateFunction(func);
	admin_cert_info_ctrl_->SetGenerateAcmeCertificateFunction(std::move(func));
}

void SettingsDialog::SetUploadCertificateFunction(CertInfoEditor::UploadCertificateFunc func)
{
	ftp_cert_info_ctrl_->SetUploadCertificateFunction(func);
	admin_cert_info_ctrl_->SetUploadCertificateFunction(std::move(func));
}

void SettingsDialog::SetGenerateAcmeAccountFunction(GenerateAcmeAccountFunc func)
{
	acme_generate_account_func_ = std::move(func);
}

void SettingsDialog::SetCertificateInfo(CertInfoEditor::cert_id id, fz::securable_socket::cert_info &&info, const fz::securable_socket::cert_info::extra &extra)
{
	wxCHECK_RET(id == CertInfoEditor::ftp_cert_id || id == CertInfoEditor::admin_cert_id, "Unknown certificate ID");

	if (id == CertInfoEditor::ftp_cert_id) {
		if (info) {
			ftp_options_.sessions().tls.cert = std::move(info);
			ftp_tls_extra_info_ = extra;
		}

		ftp_cert_info_ctrl_->SetValue(&ftp_options_.sessions().tls.cert, ftp_tls_extra_info_);
	}
	else
	if (id == CertInfoEditor::admin_cert_id) {
		if (info) {
			admin_options_.tls.cert = std::move(info);
			admin_tls_extra_info_ = extra;
		}

		admin_cert_info_ctrl_->SetValue(&admin_options_.tls.cert, admin_tls_extra_info_);
	}
}

bool SettingsDialog::Validate()
{
	server_listeners_editor_->SetAddressAndPortValidator([this](const wxString &address, unsigned int port) {
		if (auto m = admin_listeners_editor_->GetMatchingAddress(address, port, any_is_equivalent_); !m.empty())
			return address_and_port_in_other_editor(address, port, m, wxGetContainingPageTitle(admin_listeners_editor_));

		fz::hostaddress h1(address.ToStdWstring(), fz::hostaddress::format::ipvx, port);

		unsigned int admin_port = 0;
		admin_port_ctrl_->Get(admin_port);

		for (std::string_view a: { "127.0.0.1", "::1" }) {
			fz::hostaddress h2(a, fz::hostaddress::format::ipvx, admin_port);
			if (h1.equivalent_to(h2, any_is_equivalent_))
				return address_and_port_in_other_editor(address, port, wxEmptyString, wxGetContainingPageTitle(admin_listeners_editor_));
		}

		return wxString();
	});

	auto res = wxDialogEx::Validate();

	server_listeners_editor_->SetAddressAndPortValidator(nullptr);

	return res;
}

bool SettingsDialog::TransferDataFromWindow()
{
	if (!wxDialogEx::TransferDataFromWindow())
		return false;

	ftp_options_.sessions().tls.min_tls_ver = fz::tls_ver(ftp_tls_min_ver_ctrl_->GetSelection() + int(fz::tls_ver::v1_2));
	admin_options_.tls.min_tls_ver = fz::tls_ver(admin_tls_min_ver_ctrl_->GetSelection() + int(fz::tls_ver::v1_2));

	if (log_level_old_selection_ != log_level_ctrl_->GetSelection()) {
		log_level_old_selection_ = log_level_ctrl_->GetSelection();
		logger_opts_.enabled_types(GetLogTypes());
	}

	check_expired_certs();

	if (apply_function_)
		return apply_function_();

	return true;
}

wxBookCtrlBase *SettingsDialog::CreateBookCtrl()
{
	return wxCreate<wxTreebook>(this, wxID_ANY);
}

void SettingsDialog::AddBookCtrl(wxSizer *sizer)
{
	sizer->Add(GetBookCtrl(), wxSizerFlags(1).Expand().Border(wxALL, wxDlg2Px(this, wxDefaultPadding)));
}

void SettingsDialog::SetLogTypes(fz::logmsg::type types)
{
	int level = 0;

	if (fz::logmsg::debug_debug <= types)
		level = 5;
	else
	if (fz::logmsg::debug_verbose <= types)
		level = 4;
	else
	if (fz::logmsg::debug_info <= types)
		level = 3;
	else
	if (fz::logmsg::debug_warning <= types)
		level = 2;
	else
	if (types != fz::logmsg::type{})
		level = 1;

	log_level_ctrl_->SetSelection(level);
}

fz::logmsg::type SettingsDialog::GetLogTypes()
{
	std::make_unsigned_t<fz::logmsg::type> types{};

	int level = log_level_ctrl_->GetSelection();

	if (5 <= level)
		types |= fz::logmsg::debug_debug;

	if (4 <= level)
		types |= fz::logmsg::debug_verbose;

	if (3 <= level)
		types |= fz::logmsg::debug_info;

	if (2 <= level)
		types |= fz::logmsg::debug_warning;

	if (1 <= level)
		types |= fz::logmsg::default_types;

	return fz::logmsg::type(types);
}


void SettingsDialog::OnGroupsListChanging(GroupsList::Event &ev)
{
	// If adding a new group, nothing must be done here.
	if (ev.GetOldName().empty())
		return;

	std::string new_name = ev.GetNewName();
	bool questions_asked = false;
	std::string group_to_move_user_to;
	bool users_have_been_erased = true;

	enum state_e {
		delete_group = 1,
		delete_user_if_it_has_no_groups_left,
		rename_group,
		move_to_group,
		abort
	} state{};

	auto ask_user_what_to_do_on_deletion = [&] () -> state_e {
		std::vector<std::string_view> alternative_groups;
		alternative_groups.reserve(groups_.size()-1);
		for (const auto &g: groups_) {
			if (g.first != ev.GetOldName())
				alternative_groups.push_back(g.first);
		}

		wxArrayString choices;
		choices.reserve(groups_.size()+1);
		choices.push_back(wxString::Format(_S("Delete the group \"%s\""), ev.GetOldName()));
		choices.push_back(wxString::Format(_S("Delete all users that only belong to the group \"%s\""), ev.GetOldName()));

		for (const auto &g: alternative_groups) {
			choices.push_back(wxString::Format(_S("Move users to group \"%s\""), fz::to_wxString(g)));
		}

		// Creating a new window makes the old widget lose focus. We don't want that.
		wxAutoRefocuser refocus;

		int idx = wxGetSingleChoiceIndex(
			wxString::Format(_S("Some users still belong to the group \"%s\".\n\n"
								"Hit CANCEL to block the operation,\n"
								"or choose an option from the list."), ev.GetOldName()),
			wxString::Format(_S("Are you sure you want to delete group \"%s\"?"), ev.GetOldName()),
			choices,
			0
		);

		if (idx < 0)
			return abort;

		if (idx == 0)
			return delete_group;

		if (idx == 1)
			return delete_user_if_it_has_no_groups_left;

		new_name = alternative_groups[std::size_t(idx-2)];
		return move_to_group;
	};

	if (!new_name.empty())
		state = rename_group;
	else
		state = delete_group;

	for (auto uit = users_.begin(), uend = users_.end(); uit != uend;) {
		bool erase_user = false;

		auto &groups = uit->second.groups;

		auto group_name_it = std::find(groups.begin(), groups.end(), ev.GetOldName());
		if (group_name_it != groups.end()) {
			check_what_to_do:

			switch (state) {
				// Bare renaming is allowed without much fuss
				case rename_group:
					*group_name_it = new_name;
					break;

				// Deleting, on the other hand, must preserve data consistency and we need to ask the user a few questions about it.
				case delete_group:
					if (!questions_asked) {
						state = ask_user_what_to_do_on_deletion();
						questions_asked = true;
					}

					if (state != delete_group)
						goto check_what_to_do;

					groups.erase(group_name_it);
					break;

				case abort:
					ev.Veto();
					return;

				case delete_user_if_it_has_no_groups_left:
					if (groups.size() == 1)
						erase_user = true;
					else
						groups.erase(group_name_it);

					break;

				// Moving is kind of like renaming, but we also need to make sure to not create duplicates: perhaps the user is already in that group!
				case move_to_group:
					if (std::find(groups.begin(), groups.end(), new_name) == groups.end())
						*group_name_it = new_name;
					break;
			}
		}

		if (erase_user) {
			uit = users_.erase(uit);
			users_have_been_erased = true;
		}
		else {
			++uit;
		}
	}

	if (users_have_been_erased)
		users_editor_->SetUsers(users_, server_name_);
}

void SettingsDialog::check_expired_certs()
{
	std::vector<wxString> protocols_with_expired_certs;

	if (ftp_tls_extra_info_.expired())
		protocols_with_expired_certs.push_back(wxT("FTP"));

	if (admin_tls_extra_info_.expired())
		protocols_with_expired_certs.push_back(_S("Administration"));

	if (!protocols_with_expired_certs.empty())
		wxMsgError(_S("The TLS certificates for the following protocols have expired: %s."), fz::join(protocols_with_expired_certs, wxT(", ")));
}

