#include "notifier.hpp"
#include "debug.hpp"
#include "../../filezilla/mpl/for_each.hpp"

administrator::notifier::notifier(administrator &administrator, fz::ftp::session::id id, const fz::datetime &start, const std::string &peer_ip, fz::address_type peer_address_type, fz::logger_interface &logger)
	: administrator_(&administrator)
	, session_id_(id)
	, start_(start)
	, monotonic_start_(fz::monotonic_clock::now())
	, peer_ip_(peer_ip)
	, peer_address_type_(peer_address_type)
	, log_forwarder_(logger, administrator, id)
{
	auto num_of_sessions = administrator_->admin_server_.get_number_of_sessions();

	ADMINISTRATOR_DEBUG_LOG(L"%s - ns: %d", __PRETTY_FUNCTION__, num_of_sessions);

	if (num_of_sessions > 0)
		administrator_->admin_server_.broadcast<administration::session::start>(session_id_, start_, peer_ip_, peer_address_type_);
}

administrator::notifier::~notifier()
{
	fz::scoped_lock lock(mutex_);

	if (!administrator_)
		return;

	auto num_of_sessions = administrator_->admin_server_.get_number_of_sessions();

	ADMINISTRATOR_DEBUG_LOG(L"%s - ns: %d", __PRETTY_FUNCTION__, num_of_sessions);

	if (num_of_sessions > 0)
		administrator_->admin_server_.broadcast<administration::session::stop>(session_id_, fz::monotonic_clock::now()-monotonic_start_);
}

void administrator::notifier::notify_user_name(std::string_view name)
{
	fz::scoped_lock lock(mutex_);

	if (!administrator_)
		return;

	auto num_of_sessions = administrator_->admin_server_.get_number_of_sessions();

	ADMINISTRATOR_DEBUG_LOG(L"%s - ns: %d", __PRETTY_FUNCTION__, num_of_sessions);

	user_name_ = name;
	user_name_set_time_ = fz::monotonic_clock::now()-monotonic_start_;

	if (num_of_sessions > 0)
		administrator_->admin_server_.broadcast<administration::session::user_name>(session_id_, user_name_set_time_, name);
}

void administrator::notifier::notify_entry_info(std::string_view path, fz::file::mode mode, int64_t size, int64_t resume_offset)
{
	fz::scoped_lock lock(mutex_);

	if (!administrator_)
		return;

	auto num_of_sessions = administrator_->admin_server_.get_number_of_sessions();

	ADMINISTRATOR_DEBUG_LOG(L"%s - ns: %d", __PRETTY_FUNCTION__, num_of_sessions);

	entry_path_ = path;
	entry_mode_ = mode;
	entry_size_ = size;
	entry_resume_offset_ = resume_offset;

	entry_info_set_time_ = fz::monotonic_clock::now()-monotonic_start_;

	amount_at_transfer_start_ = mode == fz::file::reading ? progress_amounts_[size_t(progress_type::sent_data)] : progress_amounts_[size_t(progress_type::received_data)];

	if (num_of_sessions > 0)
		administrator_->admin_server_.broadcast<administration::session::entry_info>(session_id_, entry_info_set_time_, path, mode, size, resume_offset, amount_at_transfer_start_);
}

void administrator::notifier::notify_entry_transfer_status(entry_transfer_status status, int error)
{
	fz::scoped_lock lock(mutex_);

	if (!administrator_)
		return;

	auto num_of_sessions = administrator_->admin_server_.get_number_of_sessions();

	ADMINISTRATOR_DEBUG_LOG(L"%s - ns: %d", __PRETTY_FUNCTION__, num_of_sessions);

	auto now = fz::monotonic_clock::now()-monotonic_start_;

	if (status == entry_transfer_status::started)
		entry_transfer_started_time_ = now;
	else
	if (status == entry_transfer_status::stopped) {
		entry_path_.clear();
		entry_transfer_started_time_ = {};
	}

	if (num_of_sessions > 0)
		administrator_->admin_server_.broadcast<administration::session::entry_transfer>(session_id_, now, status, error);
}

void administrator::notifier::notify_progress(progress_type type, int64_t amount)
{
	fz::scoped_lock lock(mutex_);

	if (!administrator_)
		return;

	auto num_of_sessions = administrator_->admin_server_.get_number_of_sessions();

	ADMINISTRATOR_DEBUG_LOG(L"%s - ns: %d", __PRETTY_FUNCTION__, num_of_sessions);

	progress_amounts_[size_t(type)] = amount;

	if (num_of_sessions > 0)
		administrator_->admin_server_.broadcast<administration::session::progress>(session_id_,  fz::monotonic_clock::now()-monotonic_start_, type, amount);
}

void administrator::notifier::notify_protocol_info(const protocol_info &info)
{
	fz::scoped_lock lock(mutex_);

	if (!administrator_)
		return;

	fz::mpl::for_each<administration::session::any_protocol_info::variant>([&](auto i) {
		using T = typename decltype(i)::type;

		if (auto v = dynamic_cast<const T*>(&info)) {
			proto_info_set_time_ = fz::monotonic_clock::now()-monotonic_start_;
			proto_info_ = *v;

			return false;
		}

		return true;
	});

	auto num_of_sessions = administrator_->admin_server_.get_number_of_sessions();

	ADMINISTRATOR_DEBUG_LOG(L"%s - ns: %d", __PRETTY_FUNCTION__, num_of_sessions);

	if (num_of_sessions > 0 && proto_info_)
		administrator_->admin_server_.broadcast<administration::session::protocol_info>(session_id_, proto_info_set_time_, *proto_info_);
}


fz::logger_interface &administrator::notifier::logger()
{
	return log_forwarder_;
}

void administrator::notifier::send_session_info(administration::engine::session &session) const
{
	fz::scoped_lock lock(mutex_);

	if (administrator_->admin_server_.get_number_of_sessions() == 0)
		return;

	session.send<administration::session::start>(session_id_, start_, peer_ip_, peer_address_type_);
	session.send<administration::session::user_name>(session_id_, user_name_set_time_, user_name_);

	if (proto_info_)
		session.send<administration::session::protocol_info>(session_id_, proto_info_set_time_,*proto_info_);

	if (!entry_path_.empty()) {
		session.send<administration::session::entry_info>(session_id_, entry_info_set_time_, entry_path_, entry_mode_, entry_size_, entry_resume_offset_, amount_at_transfer_start_);

		if (entry_transfer_started_time_) {
			session.send<administration::session::entry_transfer>(session_id_, entry_transfer_started_time_, entry_transfer_status::started, 0);

			auto type = entry_mode_ == fz::file::reading ? progress_type::sent_data : progress_type::received_data;
			auto amount = progress_amounts_[size_t(type)];

			session.send<administration::session::progress>(session_id_, fz::monotonic_clock::now()-monotonic_start_, type, amount);
		}
	}
}

void administrator::notifier::detach_from_administrator()
{
	fz::scoped_lock lock(mutex_);

	administrator_ = nullptr;

	log_forwarder_.detach_from_administrator();
}

administrator::log_forwarder::log_forwarder(administrator &administrator, fz::tcp::session::id session_id)
	: administrator_(&administrator)
	, session_id_(session_id)
{
}

administrator::log_forwarder::log_forwarder(fz::logger_interface &parent, administrator &administrator, fz::tcp::session::id session_id)
	: modularized(parent)
	, administrator_(&administrator)
	, session_id_(session_id)
{
}

void administrator::log_forwarder::do_log(fz::logmsg::type t, const info_list &l, std::wstring &&message)
{
	{
		fz::scoped_lock lock(mutex_);

		if (administrator_ && administrator_->admin_server_.get_number_of_sessions() > 0)
			administrator_->admin_server_.broadcast<administration::log>(fz::datetime::now(), session_id_, t, l.as_string, message);
	}

	modularized::do_log(t, l, std::move(message));
}

void administrator::log_forwarder::detach_from_administrator()
{
	fz::scoped_lock lock(mutex_);

	administrator_ = nullptr;
}
